package org.qtunes.auth.spi;

import org.qtunes.core.*;
import java.net.*;
import java.util.*;

/**
 * A Session is a single login session and one must be in place for most {@link ControlMethod}
 * objects to be run. Each Session has a unique ID, a record of the Service that created it
 * and a number of "allowances", which determine what the Session can do. These can be checked
 * by calling the {@link #isAllowed} method.
 * </p><p>
 * When the session is created a set of Properties can be specified, which can be
 * retrieved later with the {@link #getProperties} method. Please see the API docs for that method
 * to get a list of standard properties.
 */
public class SessionImpl extends Session {

    private final AuthImpl auth;
    private final int id, ttl, created;
    private final Service service;
    private int touched;
    private Map properties;

    SessionImpl(AuthImpl auth, int id, Service service, Map<?,?> properties, int ttl) {
        this.auth = auth;
        this.service = service;
        this.id = id;
        this.ttl = ttl;
        touch();
        this.created = this.touched;
        this.properties = properties.isEmpty() ? Collections.EMPTY_MAP : Collections.unmodifiableMap(new HashMap<Object,Object>(properties));
    }

    public int getId() {
        return id;
    }

    public Service getService() {
        return service;
    }

    public int getCreated() {
        return created;
    }

    public void touch() {
        touched = (int)(System.currentTimeMillis() / 1000);
    }

    public int getLastActivity() {
        return touched;
    }

    public int getExpiryTime() {
        return ttl <= 0 ? Integer.MAX_VALUE : touched + ttl;
    }

    public boolean isAllowed(String test) {
        Collection<String> allowances = auth.getAllowances(this);
        return "".equals(test) || (!allowances.isEmpty() && "*".equals(test)) || allowances.contains(test) || allowances.contains("*");
    }

    public Collection<String> getAllowances() {
        return auth.getAllowances(this);
    }

    public String toString() {
        Collection<String> allowances = auth.getAllowances(this);
        int now = (int)(System.currentTimeMillis() / 1000);
        int ttl = getExpiryTime() - now;
        int age = now - getCreated();
        int idle = now - getLastActivity();
//        return "{id:"+id+" allow:"+allowances+", age:"+age+", ttl:"+ttl+", idle:"+idle+", bytes:"+getTraffic()+" props:"+properties+"}";
        return "{id:"+id+"}";
    }

    public Map getProperties() {
        return properties;
    }

}
