package org.qtunes.console;

import org.qtunes.core.*;
import org.qtunes.auth.*;
import org.qtunes.auth.control.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.net.*;

/**
 * A Simple controller that uses the System.in and System.out to read
 * commands. It accepts "help" to list available commands, and the
 * commands are formatted as you would call a method in JavaScript,
 * eg <code>qtunes.quit()</code>, <code>AppleServer.addRemote("*", "8192")</code>
 * or even <code>db.getFields([1,2,3,4],["Artist","Title","Album"])</code>
 */
public class ConsoleController implements Service, Runnable {

    private ServiceContext context;
    private Session session;
    private ControlMethodInvoker invoker;
    private ParamParser parser;
    private boolean cancelled;
    private Thread thread;

    protected BufferedReader getReader() {
        try {
            return new BufferedReader(System.console().reader());
        } catch (Throwable e) {
            return new BufferedReader(new InputStreamReader(System.in));
        }
    }

    protected PrintWriter getWriter() {
        try {
            return System.console().writer();
        } catch (Throwable e) {
            return new PrintWriter(System.out);
        }
    }

    public void startService(ServiceContext context) {
        this.context = context;
        Auth auth = context.getService(Auth.class);
        Map<String,Object> props = new HashMap<String,Object>();
        props.put("ttl", "0");
        props.put("service", this);
        session = auth.login(props);
        parser = new ConsoleParser();
        invoker = new ControlMethodInvoker(context);
        cancelled = false;
        thread = new Thread(this, "qTunes-"+context.getServiceName());
        thread.start();
    }

    public void stopService(ServiceContext context) {
        cancelled = true;
        thread.interrupt();
    }

    public ServiceContext getContext() {
        return context;
    }

    public void run() {
        BufferedReader r = getReader();
        PrintWriter w = getWriter();
        Session.setCurrentSession(session);

        w.write(context.getSoftwareName()+"\n");
        w.write("Type \"help\" for help\n");
        while (!cancelled) {
            w.flush();
            w.write("qtunes> ");
            w.flush();
            String s = null;
            try {
                s = r.readLine();
            } catch (IOException e) { }
            if (s!=null) {
                s = s.trim();
                if (s.length() > 0) {
                    String response;
                    try {
                        if (s.equals("quit") || s.equals("exit")) {
                            s = "qtunes.quit()";
                        }
                        if (s.equals("help")) {
                           response = help(null);
                        } else if (s.startsWith("help ")) {
                           response = help(s.substring(5));
                        } else {
                            parser.parse(s);
                            String objectname = parser.getObjectName();
                            List<Call> calls = parser.getMethodCalls();
                            response = parser.toString(invoker.invoke(context, objectname, calls));
                        }
                    } catch (Throwable e) {
                        response = parser.toString(e);
                    }
                    if (response!=null) {
                        w.write(response+"\n");
                    }
                }
            }
        }
        try { r.close(); } catch (Throwable e) {}
        try { w.close(); } catch (Throwable e) {}
    }

    /**
     * Return a String of help messages for the specified Object name
     * @param objectname the Object name, or null for general help
     */
    private String help(String objectname) {
        Object object = null;
        if (objectname!=null) {
            object = invoker.getObject(objectname);
        }
        if (object==null) {
            StringBuffer sb = new StringBuffer();
            sb.append("Type help \"name\", where 'name' is one of: qtunes, db, auth, ...");
//            String[] servicenames = context.getServiceNames();
//            for (int i=0;i<servicenames.length;i++) {
//                sb.append(", "+servicenames[i]);
//            }
            return sb.toString();
        } else {
            Method[] methods = object.getClass().getMethods();
            StringBuffer sb = new StringBuffer();
            for (int i=0;i<methods.length;i++) {
                if (methods[i].isAnnotationPresent(ControlMethod.class)) {
                    ControlMethod annot = (ControlMethod)methods[i].getAnnotation(ControlMethod.class);
                    Class[] paramtypes = methods[i].getParameterTypes();
                    String[] params = annot.params();
                    sb.append(objectname+"."+methods[i].getName()+"(");
                    for (int j=0;j<params.length;j++) {
                        if (j > 0) {
                            sb.append(", ");
                        }
                        String classname = paramtypes[j].getName();
                        classname = classname.substring(classname.lastIndexOf(".")+1);
                        sb.append(classname+" "+params[j]);
                    }
                    sb.append(")\n    ");
                    sb.append(annot.help()+"\n");
                }
            }
            return sb.toString();
        }
    }

    public Map<String,Object> reportState() {
        return null;
    }

}
