package org.qtunes.core.util;

import javax.imageio.*;
import java.io.*;
import java.awt.*;
import java.lang.ref.*;
import java.util.*;
import java.util.List;
import java.security.*;
import java.math.*;
import java.nio.*;
import java.nio.channels.*;
import java.awt.geom.*;
import java.awt.image.*;

/**
 * Utility class to scale an image to no larger than a specified size. Will
 * cache images based on the checksum of their bytes, so that identical
 * requests don't process the image more than once.
 */
public class ImageScaler {

    private static ImageScaler instance;
    private Map<BigInteger,ImageInfo> cache = new HashMap<BigInteger,ImageInfo>();
    private MessageDigest digest;
    private static final int TOLERANCE = 4;

    private ImageScaler() {
        cache = new HashMap<BigInteger,ImageInfo>();
        try {
            digest = MessageDigest.getInstance("SHA-1");
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static synchronized ImageScaler getInstance() {
        if (instance == null) {
            instance = new ImageScaler();
        }
        return instance;
    }

    /**
     * Return a ByteBuffer containing rendered image. The image
     * will have a maximum dimension of maxwidth x maxheight - if
     * the image supplied in "in" has to be scaled to meet that,
     * the returned image will be in the requested format, otherwise
     * the original image is returned.
     * If the image can't be parsed or is null, this method returns
     * null
     *
     * @param in a ByteBuffer containing a rendered Image, or null
     * @param maxwidth the maximum width in pixels
     * @param maxheight the maximum width in pixels
     * @param format if the image has to be scaled, what format to create the saved image as
     * @return a ByteBuffer containing a bitmap image no larger than the requestest size, or null
     */
    public synchronized ByteBuffer getScaledImage(ByteBuffer in, int maxwidth, int maxheight, String format) throws IOException {
        if (in == null || maxwidth == 0 || maxheight == 0) {
            return null;
        } else {
            BigInteger key = checksum(in);
            ImageInfo info = cache.get(key);
            ScaledImage si = null;
            if (info == null) {
                info = new ImageInfo();
                cache.put(key, info);
                si = createImage(in, maxwidth, maxheight, format, info);
            }
            if (info.error) {
                return null;
            } else if (info.width <= maxwidth && info.height <= maxheight) {
                return in;
            } else {
                if (si == null) {
                    for (Iterator<SoftReference<ScaledImage>> i = info.scaled.iterator();i.hasNext();) {
                        ScaledImage tsi = i.next().get();
                        if (tsi == null) {
                            i.remove();
                        } else if (tsi.width <= maxwidth && (si==null || tsi.width > si.width) && tsi.height <= maxheight && (si==null || tsi.height > si.height)) {
                            si = tsi;
                        }
                    }
                }
                if (si == null || maxwidth-si.width > TOLERANCE || maxheight-si.height > TOLERANCE) {
                    si = createImage(in, maxwidth, maxheight, format, info);
                }
                si.buf.position(0);
                return si.buf;
            }
        }
    }

    private ScaledImage createImage(ByteBuffer buf, int maxwidth, int maxheight, String format, ImageInfo info) throws IOException {
        buf.position(0);
        InputStream in = createInputStream(buf);
        BufferedImage img = ImageIO.read(in);
        in.close();
        if (img != null) {
            info.width = img.getWidth();
            info.height = img.getHeight();
            if (img.getWidth() > maxwidth || img.getHeight() > maxheight) {
                // Yes, must loop to get decent output. No, can't use AffineImageOp
                // as we must use TYPE_INT_RGB - any alpha gives black JPEG images 
                while (img.getWidth() > maxwidth || img.getHeight() > maxheight) {
                    double scale = Math.max(0.5, Math.min((double)maxwidth / img.getWidth(), (double)maxheight / img.getHeight()));

                    int newwidth = (int)Math.round(img.getWidth() * scale);
                    int newheight = (int)Math.round(img.getHeight() * scale);
                    BufferedImage outimg = new BufferedImage(newwidth, newheight, BufferedImage.TYPE_INT_RGB);
                    Graphics2D g = outimg.createGraphics();
                    g.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
                    g.drawRenderedImage(img, AffineTransform.getScaleInstance(scale, scale));
                    g.dispose();
                    img = outimg;
                }
                ByteArrayOutputStream out = new ByteArrayOutputStream();
                ImageIO.write(img, format, out);
                ScaledImage si = new ScaledImage(img.getWidth(), img.getHeight(), ByteBuffer.wrap(out.toByteArray()));
                info.scaled.add(new SoftReference<ScaledImage>(si));
                return si;
            }
        } else {
            info.error = true;
        }
        return null;
    }

    // Why can't ImageIO.createImageInputStream() take a ByteBuffer?
    private static InputStream createInputStream(final ByteBuffer buffer) {
        return new BufferedInputStream(new InputStream() {    // Buffer: bug 6347575
            public int read() {
                return buffer.hasRemaining() ? buffer.get() & 0xFF : -1;
            }
            public int read(byte[] buf) {
                return read(buf, 0, buf.length);
            }
            public int read(byte[] buf, int off, int len) {
                if (!buffer.hasRemaining()) {
                    return -1;
                }
                int pos = buffer.position();
                buffer.get(buf, off, Math.min(len, buffer.remaining()));
                return buffer.position() - pos;
            }
        });
    }

    private static class ImageInfo {
        int width, height;
        boolean error;
        List<SoftReference<ScaledImage>> scaled = new LinkedList<SoftReference<ScaledImage>>();
    }

    private static class ScaledImage {
        final int width, height;
        final ByteBuffer buf;
        ScaledImage(int width, int height, ByteBuffer buf) {
            this.width = width;
            this.height = height;
            this.buf = buf;
        }
    }

    private final BigInteger checksum(ByteBuffer buf) throws IOException {
        buf.position(0);
        DigestOutputStream out = new DigestOutputStream(new OutputStream() {
            public void write(int b) {}
            public void write(byte[] b, int off, int len) {}
        }, digest);
        WritableByteChannel channel = Channels.newChannel(out);
        channel.write(buf);
        channel.close();
        byte[] data = digest.digest();
        return new BigInteger(data);
    }

}
