package org.qtunes.daap;

import org.qtunes.core.*;
import org.qtunes.db.*;
import org.qtunes.web.WebConnection;
import java.io.*;
import java.util.*;
import java.lang.ref.SoftReference;
import java.util.zip.*;
import java.net.*;

/**
 * The base class for all request handlers, these are called by the RootHandler
 * depending on the path. Subclasses that return a "Block" need to override handleDAAP
 */
abstract class AbstractHandler {

    private static Map<String,SoftReference<Response>> cache = new HashMap<String,SoftReference<Response>>();
    private int cacherevision;

    /**
     * Return true if this request should fail if there's no session
     */
    boolean isSessionRequired() {
        return true;
    }

    /**
     * Return true if this request requires a password (if the server requires one at all)
     */
    boolean isPasswordRequired() {
        return true;
    }

    /**
     * Return a key that uniquely identifies this request - a subsequent request with the
     * same key would be served with the cached copy. A value of null means the response
     * is not cached.
     */
    String getCacheKey(WebConnection con) {
        return null;
    }

    /**
     * Return true if this request should set the Content-Type header
     */
    boolean isStandardContentType() {
        return true;
    }

    void handle(DaapServer daap, WebConnection con) throws IOException {
        ServiceContext context = daap.getServiceContext();
        context.debug("<-"+con.getRequestURI().toString());

        if (isStandardContentType()) {
            con.addResponseHeader("Content-Type", "application/x-dmap-tagged");
        }
        boolean gzip = con.getRequestHeader("Accept-Encoding")!=null && con.getRequestHeader("Accept-Encoding").indexOf("gzip") >= 0;
        Database db = daap.getDatabase();

        Response response = null;
        String cachekey = null; // getCacheKey(con);
        if (cachekey!=null && cacherevision==db.getRevision() && (response=cacheGet(cachekey))!=null) {
            context.debug("->(cached response)");
        } else {
            Block block = handleDAAP(daap, con);
            if (con.getResponseCode()==-1) {
                response = new Response(block);
                context.debug("->"+block.dump());
                if (cachekey!=null) {
                    cachePut(cachekey, response);
                    cacherevision = db.getRevision();
                }
            }
        }

        if (response!=null) {
            byte[] data;
            if (gzip) {
                con.addResponseHeader("Content-Encoding", "gzip");
                data = response.getCompressedData();
            } else {
                data = response.getUncompressedData();
            }
            con.sendResponseHeaders(200, data.length);
            OutputStream out = con.getOutputStream();
            out.write(data);
            out.close();
        } else {
            context.debug("->"+con.getResponseCode());
        }
    }

    abstract Block handleDAAP(DaapServer daap, WebConnection con) throws IOException;

    private final class Response {
        private Block block;
        private byte[] compresseddata;
        private byte[] uncompresseddata;

        Response(Block block) {
            this.block = block;
        }

        synchronized byte[] getCompressedData() throws IOException {
            if (compresseddata==null) {
                ByteArrayOutputStream bout = new ByteArrayOutputStream();
                OutputStream out = new GZIPOutputStream(bout);
                if (block!=null) {
                    block.write(out);
                    block = null;
                } else {
                    out.write(uncompresseddata);
                }
                out.close();
                compresseddata = bout.toByteArray();
            }
            return compresseddata;
        }

        synchronized byte[] getUncompressedData() throws IOException {
            if (uncompresseddata==null) {
                ByteArrayOutputStream bout = new ByteArrayOutputStream();
                if (block!=null) {
                    block.write(bout);
                    block = null;
                } else {
                    byte[] temp = new byte[8192];
                    int l;
                    InputStream in = new GZIPInputStream(new ByteArrayInputStream(compresseddata));
                    while ((l=in.read(temp, 0, temp.length))>=0) {
                        bout.write(temp, 0, l);
                    }
                }
                uncompresseddata = bout.toByteArray();
            }
            return uncompresseddata;
        }
    }

    private static synchronized Response cacheGet(String key) {
        SoftReference<Response> ref = cache.get(key);
        return ref==null ? null : ref.get();
    }

    private static synchronized void cachePut(String key, Response response) {
        cache.put(key, new SoftReference<Response>(response));
    }

}
