package org.qtunes.daap;

import org.qtunes.db.*;
import java.security.*;

/**
 * The class defines some DerivedField objects which are needed by the Apple
 * service
 */
class DaapField {

    /**
     * The AlbumID - a unique number for each album. Required to differentiate
     * between two different albums called "Greatest Hits", for instance, it's
     * a unique long.
     */
    static final DerivedField<Long> AlbumId = new DerivedField<Long>("AlbumId") {
        public Long get(Track track) {
            String album = (String)track.get(Field.Album);
            String artist = (String)track.get(GroupArtist);
            if (album!=null && artist!=null) {
                long val = digest64(album, artist);
                if (val >= 0) {
                    return new Long(val);
                }
            }
            return new Long(0);
        }

        public boolean isCacheable() {
            return true;
        }
    };

    /**
     * The Group Artist - the artist for the selected album.
     */
    static final DerivedField<String> GroupArtist = new DerivedField<String>("GroupArtist") {
        public String get(Track track) {
            String artist;
            boolean compilation = track.get(Field.Compilation);
            if (compilation) {
                artist = "Various";
            } else {
                artist = (String)track.get(Field.AlbumArtist);
                if (artist==null) {
                    artist = (String)track.get(Field.Artist);
                }
            }
            return artist;
        }
    };

    static final AbstractField<String> MediaDescription = new DerivedField<String>("MediaDescription") {
        public String get(Track track) {
            return "MPEG audio file";
        }
    };

    /**
     * A unique number for each track
     */
    static final AbstractField<Integer> DatabaseIndex = new DerivedField<Integer>("DatabaseIndex") {
        public Integer get(Track track) {
            return Integer.valueOf(track.getIndex());
        }
    };

    static final AbstractField<Integer> ItemKind = new DerivedField<Integer>("ItemKind") {
        public Integer get(Track track) {
            return Integer.valueOf(2);
        }
    };

    static final AbstractField<Integer> SongDataKind = new DerivedField<Integer>("SongDataKind") {
        public Integer get(Track track) {
            return Integer.valueOf(0);
        }
    };

    static final AbstractField<Integer> MediaKind = new DerivedField<Integer>("MediaKind") {
        public Integer get(Track track) {
            return Integer.valueOf(1);
        }
    };

    private static MessageDigest digest;

    /** 
     * Return a 64-bit hash code, probably unique, of the specified keys
     * Synchronized because we're sharing a digest, but could put this
     * into a ThreadLocal if it ever causes contention (unlikely)
     */
    synchronized static long digest64(String... keys) {
        try {
            if (digest==null) {
                digest = MessageDigest.getInstance("MD5");
            }
            for (int i=0;i<keys.length;i++) {
                if (keys[i]!=null) {
                    digest.update(keys[i].getBytes("ISO-8859-1"));
                }
            }
            byte[] value = digest.digest();
            long v = ((value[ 0]^value[ 1])<<54) ^ ((value[2]^value[3])<<48) ^ 
                     ((value[ 4]^value[ 5])<<40) ^ ((value[6]^value[7])<<32) ^ 
                     ((value[ 8]^value[ 9])<<24) ^ ((value[10]^value[11])<<16) ^ 
                     ((value[12]^value[13])<<8)  ^ ((value[14]^value[15])<<0);
            v &= 0x7FFFFFFFFFFFFFFFl;
            return v;
        } catch (Exception e) {
            return -1;
        }
    }

}
