package org.qtunes.daap;

import org.qtunes.db.*;
import org.qtunes.player.Player;
import java.util.*;

/**
 * A wrapper around a  regular core.Playlist which contains extra
 * information required by the DAAP/DACP protocols. Unlike regular
 * playlists, each track in the playlist has  "container item id" or
 * "CID" for short. Any operations to add or remove tracks are done
 * by reference to this CID, rather than referencing the index in the
 * list directly.
 *
 * Some terminology:
 *
 * Base playlist - the top-level playlist, there can be only one. The "abpl" Block
 * Smart playlist - is the Playlist based on a shifting set of Tracks? The "apSP" Block
 * Special playlist - controls the appearance of the Playlist in Remote
 * Editflags - flags determining what can be done to the Playlist in Remote ("meds" Block)
 */
class DaapPlaylist {

    // Flags for the "meds" field
    private static final int MEDS_USER=0x20, MEDS_PURGE=0x10, MEDS_REBUILD=0x8, MEDS_SORT=0x4, MEDS_DEL=0x2, MEDS_ADD=0x1;
    private final String name;
    private final List<Entry> tracks;           // List of [cid,track] tuples
    private List<Entry> copy;                   // temp copy of same, for moving
    private final Playlist playlist;
    private boolean base, smart;
    private int special, editflags;
    private long persistent;
    private int index;
    private int nextcid;                // CID of the next track to add
    private Database db;

    // Map which keeps the canonical mapping from Playlist -> DaapPlaylist,
    // as well as a couple of other useful bits, ie next index.
    private static Map<DaapServer,Map<Object,Object>> all = new HashMap<DaapServer,Map<Object,Object>>();

    private DaapPlaylist(Database db, String name, Playlist playlist) {
        this.db = db;
        this.name = name;
        this.playlist = playlist;
        List<Track> origtracks = playlist.getTracks();
        tracks = new ArrayList<Entry>(origtracks.size());
        for (int i=0;i<origtracks.size();i++) {
            tracks.add(new Entry(++nextcid, origtracks.get(i)));
        }
    }

    int getIndex() {
        return index;
    }

    long getPersistentId() {
        return persistent;
    }

    String getName() {
        return name;
    }

    Playlist getPlaylist() {
        return playlist;
    }

    /**
     * Is the Playlist the "base playlist", as determined by the aeBP Block
     */
    boolean isBasePlaylist() {
        return base;
    }

    /**
     * Is the Playlist a Smart playlist?
     */
    boolean isSmart() {
        return smart;
    }

    /**
     * Get the value of the "special" field, as returned by the "aePS" block.
     * So far we have 6 (Music), 4 (Films), 5 (TV Programmes), 13 (iTunes-U),
     * 7 (Books), 8 (Purchased), 12 (Genius), 15 (Genius Mixes), 16 (individual
     * genius mixes). This seems to primarily control the appearance flag
     */
    int getSpecial() {
        return special;
    }

    /**
     * Return the "editflags" flags (the "meds" field).
     * Analysis of the flags gives the following probable meanings:
     * Bit 1 - if set we can add tracks to the list
     * Bit 2 - if set we can delete tracks from the list
     * Bit 3 - if set we can sort the tracks in the list
     * Bit 4 - only set for Genius - we can recreate the list?
     * Bit 5 - if set we can delete the playlist
     * Bit 6 - seems to be set for user-created (as opposed to system-created) lists
     */
    int getEditFlags() {
        return editflags;
    }

    /**
     * Get the list of all CIDs in this Playlist,
     * in order
     */
    synchronized int[] getCIDs() {
        int[] ids = new int[tracks.size()];
        for (int i=0;i<ids.length;i++) {
            ids[i] = tracks.get(i).cid;
        }
        return ids;
    }

    /**
     * Given a CID, return it's position in the playlist
     * or -1 if it's not there
     */
    synchronized int getCIDIndex(int cid) {
        for (int i=0;i<tracks.size();i++) {
            if (tracks.get(i).cid==cid) {
                return i;
            }
        }
        return -1;
    }

    /**
     * Remove the specified CID from the playlist
     */
    synchronized void remove(int cid) {
        if ((editflags & MEDS_DEL)!=0) {
            int ix = getCIDIndex(cid);
            if (ix >= 0) {
               tracks.remove(ix);
               ((ManualPlaylist)playlist).remove(ix);
            }
        } else {
            throw new IllegalStateException("Can't delete from Playlist");
        }
    }

    /**
     * Move the track with the specified CID to the specified
     * position in the playlist. The values of the arguments
     * are as passed in by Remote.
     */
    synchronized void move(int fromix, int toix) {
        if (copy==null) {
            copy = new ArrayList<Entry>();
            copy.addAll(tracks);
        }
        Entry e = tracks.get(fromix - 1);
        int oldix = copy.indexOf(e);
        copy.add(toix, e);
        if (oldix >= toix) {
            oldix++;
        }
        copy.remove(oldix);
    }

    /**
     * Add the tracks from the specified Playlist to this Playlist
     */
    synchronized void addTracks(Playlist add) {
        if ((editflags & MEDS_ADD)!=0) {
            List<Track> newtracks = add.getTracks();
            for (int i=0;i<newtracks.size();i++) {
                tracks.add(new Entry(++nextcid, newtracks.get(i)));
            }
            ((ManualPlaylist)playlist).addAll(add);
        } else {
            throw new IllegalStateException("Can't add to Playlist");
        }
    }

    /**
     * Commit any remove/add/move operations to the database.
     */
    synchronized void save() {
        if (copy!=null) {
            if (playlist instanceof ManualPlaylist) {
                synchronized(playlist) {
                    ((ManualPlaylist)playlist).clear();
                    for (int i=0;i<copy.size();i++) {
                        ((ManualPlaylist)playlist).add(copy.get(i).track);
                    }
                }
            }
            tracks.clear();
            tracks.addAll(copy);
            copy = null;
        }
        db.putPlaylist(name, playlist);
    }

    public String toString() {
        return "{name:"+getName()+" ix:"+index+" pix:0x"+Long.toHexString(persistent)+" tracks:"+tracks+"}";
    }

    //--------------------------------------------------------------------------------
    // Remaining methods are static
    //--------------------------------------------------------------------------------

    /**
     * Create a new SmartPlaylist from the specified filter. The filter
     * is in the apple format, not too hard to work out. Basically converts
     * that filter to the syntax used by SmartPlaylist
     */
    static SmartPlaylist getSmartPlaylist(String filter, Database db) {
        if (filter==null) {
            return null;
        }
        StringBuffer out = new StringBuffer();
        StringBuffer keyval = null;
        for (int i=0;i<filter.length();i++) {
            char c = filter.charAt(i);
            if (c=='\\' && keyval!=null) {
                keyval.append(filter.charAt(++i));
            } else if (c=='\'' && keyval==null) {
                keyval = new StringBuffer();
            } else if (c=='\'' && keyval!=null) {
                char op = 0;
                boolean not = false;
                String key=null, val=null;
                for (int j=0;j<keyval.length();j++) {
                    c = keyval.charAt(j);
                    if (c==':' || c=='+' || c=='-' || c=='!') {
                        op = c;
                        key = keyval.substring(0, j);
                        if (c=='!') {
                            not = true;
                            op = keyval.charAt(++j);
                        }
                        val = keyval.substring(j+1);
                        break;
                    }
                }
                if (op==0) {
                    throw new IllegalArgumentException("Can't parse \""+filter+"\": failed on expression \""+keyval+"\"");
                }
                keyval = null;
                Blocks.BlockDef blockdef = Blocks.getFromDescription(key.toString());
                if (blockdef!=null && blockdef.getField()!=null) {
                    out.append(blockdef.getField().getName());
                    if (val.indexOf("*")>=0) {
                        val = val.replaceAll("\\*", ".*");
                        out.append(not ? "!~" : "=~");
                    } else if (op==':') {
                        out.append(not ? "!=" : "=");
                    } else if (op=='+') {
                        out.append(not ? "<=" : ">");
                    } else if (op=='-') {
                        out.append(not ? ">=" : "<");
                    }
                    out.append("\""+val.replaceAll("\"", "\\\\\"")+"\"");
                }
            } else if (keyval!=null) {
                keyval.append(c);
            } else if (c=='(') {
                out.append("(");
            } else if (c==')') {
                out.append(")");
            } else if (c==' ') {
                out.append(" and ");
            } else if (c=='+') {
                out.append(" and ");
            } else if (c==',') {
                out.append(" or ");
            }
        }
        return db.createSmartPlaylist(out.toString());
    }

    /**
     * Create a new SmartPlaylist from the specified filter. The filter
     * is in the apple format, not too hard to work out. Basically converts
     * that filter to the syntax used by SmartPlaylist
     */
    static Playlist getSortedSmartPlaylist(String filter, Database db, String sort) {
        SmartPlaylist pl = getSmartPlaylist(filter, db);
        if (sort==null || sort.equals("")) {
            return pl;
        } else if ("album".equalsIgnoreCase(sort)) {
            return pl.sort("+album", null);
        } else if ("artist".equalsIgnoreCase(sort)) {
            return pl.sort("+artist", null);
        } else if ("track".equalsIgnoreCase(sort)) {
            return pl.sort("+title", null);
        } else {
            throw new IllegalArgumentException("Unknown sort field \""+sort+"\"");
        }
    }

    //----------------------------------------------------------------------
    // Methods to do with managing the mapping from Playlist -> DaapPlaylist
    //----------------------------------------------------------------------

    /**
     * Remove any playlist mappings for the specified DaapServer
     */
    static synchronized void free(DaapServer daap) {
        all.remove(daap);
    }

    /**
     * Get the DaapPlaylist for the base playlist of this DaapServer
     */
    static synchronized DaapPlaylist getBasePlaylist(DaapServer daap) {
        Map<Object,Object> playlists = all.get(daap);
        if (playlists==null) {
            all.put(daap, playlists = new HashMap<Object,Object>());
        }
        DaapPlaylist apl = (DaapPlaylist)playlists.get("base");
        if (apl==null) {
            playlists.put("base", apl = new DaapPlaylist(daap.getDatabase(), daap.getServerName(), daap.getDatabase().getPlaylist("*")));
            apl.index = 1;
            apl.persistent = daap.getUniqueID() + apl.index;
            apl.base = true;
            apl.smart = true;
        }
        return apl;
    }

    /**
     * Get the DaapPlaylist for the "Music" playlist of this DaapServer
     */
    static synchronized DaapPlaylist getMusicPlaylist(DaapServer daap) {
        Map<Object,Object> playlists = all.get(daap);
        if (playlists==null) {
            all.put(daap, playlists = new HashMap<Object,Object>());
        }
        DaapPlaylist apl = (DaapPlaylist)playlists.get("Music");
        if (apl==null) {
            playlists.put("Music", apl = new DaapPlaylist(daap.getDatabase(), "Music", daap.getDatabase().getPlaylist("*")));
            apl.index = 2;
            apl.persistent = daap.getUniqueID() + apl.index;
            apl.special = 6;
            apl.smart = true;
        }
        return apl;
    }

    /**
     * Get the DaapPlaylist for the "now-playing" playlist of this DaapServer.
     * Note this is required because the Player does things differently to iTunes -
     * in the Player you have to select a Playlist or "Music" and play. Our Player
     * doesn't have that restriction, and it plays whatever is queued.
     */
    static synchronized DaapPlaylist getNowPlayingPlaylist(DaapServer daap, Player player) {
        if (player == null) {
            return null;
        } else {
            Map<Object,Object> playlists = all.get(daap);
            if (playlists==null) {
                all.put(daap, playlists = new HashMap<Object,Object>());
            }
            DaapPlaylist apl = (DaapPlaylist)playlists.get(player);
            if (apl==null) {
                playlists.put(player, apl = new DaapPlaylist(daap.getDatabase(), "Now Playing", player.getPlaylist()));
                apl.index = 3;
                apl.persistent = DaapServer.getUniqueID(player);
                apl.smart = true;
            }
            return apl;
        }
    }

    /**
     * Get the DaapPlaylist for the specified Playlist
     */
    static synchronized DaapPlaylist getInstance(DaapServer daap, String name, Playlist playlist) {
        Map<Object,Object> playlists = all.get(daap);
        if (playlists==null) {
            all.put(daap, playlists = new HashMap<Object,Object>());
        }
        DaapPlaylist apl = (DaapPlaylist)playlists.get(playlist);
        if (apl==null) {
            playlists.put(playlist, apl = new DaapPlaylist(daap.getDatabase(), name, playlist));
            Integer ix = (Integer)playlists.get("index");
            apl.index = ix==null ? 10 : ix.intValue();
            playlists.put("index", new Integer(apl.index + 1));
            apl.persistent = daap.getUniqueID() + apl.index;
            if (playlist instanceof ManualPlaylist) {
                apl.editflags = MEDS_ADD | MEDS_DEL | MEDS_SORT | MEDS_PURGE;
            } else if (playlist instanceof SmartPlaylist) {
                apl.editflags = MEDS_PURGE;
                apl.smart = true;
            }
        }
        return apl;
    }

    private static class Entry {
        final int cid;
        final Track track;

        Entry(int cid, Track track) {
            this.cid = cid;
            this.track = track;
        }

        public String toString() {
            return "{"+cid+":"+track.getIndex()+"}";
        }
    }

}
