package org.qtunes.daap;

import org.qtunes.core.*;
import org.qtunes.core.util.Base64;
import org.qtunes.db.*;
import org.qtunes.web.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.text.SimpleDateFormat;

/**
 * The base HttpHandler for all DACP and DAAP requests. Farms them
 * out depending on the path, and checks for password, session etc.
 * first.
 */
class RootHandler implements WebHandler {

    private final Map<String,AbstractHandler> map = new HashMap<String,AbstractHandler>();
    private final DaapServer daap;
    private final ServiceContext context;
    private boolean servedacp, servedaap;

    /**
     * @param servedaap whether the DAAP protocol is supported
     * @param servedacp whether the DAAP protocol is supported
     */
    RootHandler(DaapServer server) {
        this.daap = server;
        this.context = server.getServiceContext();
        map.put("/server-info", new HandlerServerInfo());
        map.put("/login", new HandlerLogin());
        map.put("/logout", new HandlerLogout());
        map.put("/databases/items", new HandlerItems());
        map.put("/ctrl-int/setspeakers", new HandlerSetProperty());
        map.put("/update", new HandlerUpdate());
        map.put("/databases", new HandlerDatabases());
        map.put("/databases/containers", new HandlerContainers());
        map.put("/databases/containers/items", new HandlerContainerItems());
        map.put("/databases/items/", new HandlerSong());
        map.put("/databases/browse/artists", new HandlerBrowse());
        map.put("/databases/browse/albums", new HandlerBrowse());
        map.put("/databases/browse/genres", new HandlerBrowse());
        map.put("/databases/browse/composers", new HandlerBrowse());
        map.put("/databases/groups", new HandlerGroups());
        map.put("/databases/groups/extra_data/artwork", new HandlerGroupArtwork());
        map.put("/databases/items/extra_data/artwork", new HandlerItemArtwork());
        map.put("/databases/edit", new HandlerEdit());
        map.put("/databases/containers/edit", new HandlerContainerEdit());
    }

    void setServeDaap(boolean servedaap) {
        if (this.servedaap != servedaap) {
            if (servedaap) {
                map.put("/content-codes", new HandlerContentCodes());
            } else {
                map.remove("/content-codes");
            }
            this.servedaap = servedaap;
        }
    }

    void setServeDacp(boolean servedacp) {
        if (this.servedacp != servedacp) {
            if (servedacp) {
                map.put("/ctrl-int", new HandlerCtrlInt());
                map.put("/ctrl-int/cue", new HandlerCue());
                map.put("/ctrl-int/playstatusupdate", new HandlerPlayStatusUpdate());
                map.put("/ctrl-int/playspec", new HandlerPlaySpec());
                map.put("/ctrl-int/getproperty", new HandlerGetProperty());
                map.put("/ctrl-int/getspeakers", new HandlerGetSpeakers());
                map.put("/ctrl-int/nextitem", new HandlerNextItem());
                map.put("/ctrl-int/nowplayingartwork", new HandlerNowPlayingArtwork());
                map.put("/ctrl-int/pause", new HandlerPause());
                map.put("/ctrl-int/playpause", new HandlerPause());
                map.put("/ctrl-int/previtem", new HandlerNextItem());
                map.put("/ctrl-int/setproperty", new HandlerSetProperty());
            } else {
                map.remove("/ctrl-int");
                map.remove("/ctrl-int/cue");
                map.remove("/ctrl-int/playstatusupdate");
                map.remove("/ctrl-int/playspec");
                map.remove("/ctrl-int/getproperty");
                map.remove("/ctrl-int/getspeakers");
                map.remove("/ctrl-int/nextitem");
                map.remove("/ctrl-int/nowplayingartwork");
                map.remove("/ctrl-int/pause");
                map.remove("/ctrl-int/playpause");
                map.remove("/ctrl-int/previtem");
                map.remove("/ctrl-int/setproperty");
            }
            this.servedacp = servedacp;
        }
    }

    public void handle(WebConnection con) throws IOException {
        URI uri = con.getRequestURI();
        Map<String,Object> params = con.getParameters();
        String path = uri.getPath();
        try {
            // Strip some parameters out of the request
            if (path.startsWith("/ctrl-int/")) {
                String database = path.substring(10, path.indexOf('/', 10));
                path = path.substring(0, 10) + path.substring(11 + database.length());
                params.put("database", database);
            } else if (path.startsWith("/databases/")) {
                String database = path.substring(11, path.indexOf('/', 11));
                path = path.substring(0, 11) + path.substring(12 + database.length());
                params.put("database", database);
            }
            if (path.startsWith("/databases/items/")) {
                int ix = path.indexOf('/', 17);
                String item = path.substring(17, ix < 0 ? path.length() : ix);
                path = path.substring(0, 17) + (ix < 0 ? "" : path.substring(18 + item.length()));
                params.put("item", item);
            } else if (path.startsWith("/databases/groups/")) {
                String group = path.substring(18, path.indexOf('/', 18));
                path = path.substring(0, 18) + path.substring(19 + group.length());
                params.put("group", group);
            } else if (path.startsWith("/databases/containers/")) {
                String container = path.substring(22, path.indexOf('/', 22));
                path = path.substring(0, 22) + path.substring(23 + container.length());
                params.put("container", container);
            }
        } catch (RuntimeException e) {
            context.error("Exception while splitting \""+path+"\"", e);
            throw e;
        }
        con.getParameters().put("__path", path);

        // Find the matching handler and execute it
        AbstractHandler handler = map.get(path);
        if (handler!=null) {
            String sessionid = con.getParameter("session-id");
            if (sessionid != null) {
                int sid = Integer.parseInt(sessionid.toString());
                Session session = daap.getAuth().getSession(sid);
                Session.setCurrentSession(session);
                if (session != null) {
                    session.touch();
                }
            }

            try {
                con.setResponseHeader("DAAP-Server", context.getSoftwareName()+"/"+daap.getContext().getServiceName()+"/"+con.getWebService().getContext().getServiceName());
                con.setResponseHeader("Date", new Date());
                if (handler.isPasswordRequired() && !checkPassword(daap.getPassword(), con.getRequestHeader("Authorization"))) {
                    // Not working - don't seem to get a second request from the client
                    // Headers come out as "Www-authenticate". Legal but causing issues?
                    con.setResponseHeader("WWW-Authenticate", "Basic realm=\""+daap.getContext().getServiceName()+"\"");
                    con.sendResponseHeaders(401, -1);
                } else if (!handler.isSessionRequired() || Session.getCurrentSession().isAllowed("*")) {
                    context.debug("Handing \""+uri.getPath()+" as \""+path+"\" with "+handler.getClass().getName());
                    handler.handle(daap, con);
                } else {
                    context.debug("Session "+Session.getCurrentSession()+" invalid for "+uri.getPath());
                    con.sendResponseHeaders(503, -1);
                }
            } catch (IOException e) {
                if (!"Broken pipe".equals(e.getMessage()) && !"stream is closed".equals(e.getMessage())) {
                    context.error(path, e);
                }
                throw e;
            } catch (RuntimeException e) {
                context.error(path, e);
                throw e;
            } catch (Error e) {
                context.error(path, e);
                throw e;
            } finally {
                con.close();
                Session.setCurrentSession(null);
            }
        } else {
            throw new Error("No handler for "+uri+" path="+path);
        }
    }

    private boolean checkPassword(String passwd, String auth) {
        if (passwd!=null) {
            if (auth!=null && auth.startsWith("Basic ")) {
                try {
                    String userpass = new String(Base64.decode(auth.substring(6)), "ISO-8859-1");
                    String testpasswd = userpass.substring(userpass.indexOf(".") + 1);
                    return testpasswd.equals(passwd);
                } catch (UnsupportedEncodingException e) {
                    throw new RuntimeException(e);
                }
            }
            return false;
        } else {
            return true;
        }
    }

}
