package org.qtunes.daap;

import org.qtunes.core.*;
import org.qtunes.player.*;
import org.qtunes.zeroconf.*;
import java.net.*;
import java.io.*;
import java.util.*;
import java.util.concurrent.*;

/**
 * Local class which handles all the ZeroConf announcements, incoming and
 * outgoing. Was part of DaapServer but too complex so spun off.
 */
class ServiceRegistrar extends Thread {

    private final ServiceContext context;
    private final DaapServer server;
    private final int port;
    private final Set<Player> players;

    private RemoteManager remotemanager;
    private volatile boolean cancelled, sharing;
    private volatile int touch;

    ServiceRegistrar(DaapServer server, InetSocketAddress address) {
        super("qTunes-"+server.getContext().getServiceName()+"-ServiceRegistrar");
        this.server = server;
        this.port = address.getPort();
        context = server.getContext();
        players = new CopyOnWriteArraySet<Player>();
    }

    void setRemoteManager(RemoteManager manager) {
        this.remotemanager = manager;
    }

    /**
     * Notify the ServiceRegistrar of a new Player. If there's a Remote control
     * for the player, advertise it's touch-able service
     */
    void addPlayer(Player player) {
        players.add(player);
        update();
    }

    /**
     * Notification of the removal of a player
     */
    void removePlayer(Player player) {
        players.remove(player);
        update();
    }


    /**
     * Notification that our DaapServer is now sharing (or not) - advertise
     * the daap service as appropriate.
     */
    void setSharing(boolean sharing) {
        if (sharing != this.sharing) {
            this.sharing = sharing;
            update();
        }
    }

    /**
     * Something has changed - update the service announcements
     */
    synchronized void update() {
        touch++;
        interrupt();
    }

    /**
     * Cancel the announcements and stop the service
     */
    synchronized void cancel() {
        touch++;
        cancelled = true;
        interrupt();
    }

    public void run() {
        int intouch = touch;
        Set<ServerState> touchablestates = new LinkedHashSet<ServerState>();
        ServerState daapstate = null;

        while (!cancelled) {
            try {
                synchronized(this) {
                    if (touch == intouch) {
                        this.wait();
                    }
                    intouch = touch;
                }
            } catch (InterruptedException e) { }

            ZeroConf zeroconf = context.getService(ZeroConf.class);
            if (zeroconf != null) {
                ServerState newstate = cancelled ? null : getDaapInstance();
                if (newstate == null ? daapstate != null : !newstate.equals(daapstate)) {
                    if (daapstate != null) {
                        context.debug("Stopping daap service \""+daapstate+"\"");
                        daapstate.unregister();
                    }
                    if (newstate != null) {
                        context.debug("Starting daap service \""+newstate+"\"");
                        newstate.register(zeroconf, "_daap._tcp");
                        daapstate = newstate;
                    }
                }

                for (Iterator<ServerState> i = touchablestates.iterator();i.hasNext();) {
                    ServerState oldstate = i.next();
                    newstate = getTouchableInstance(oldstate.player);
                    if (!oldstate.equals(newstate)) {
                        context.debug("Stopping touchable service \""+oldstate+"\"");
                        oldstate.unregister();
                        i.remove();
                    }
                }
                if (!cancelled) {
                    for (Iterator<Player> i = players.iterator();i.hasNext();) {
                        Player player = i.next();
                        if (player != null) {
                            newstate = getTouchableInstance(player);
                            if (newstate != null && touchablestates.add(newstate)) {
                                context.debug("Starting touch-able service \""+newstate+"\"");
                                newstate.register(zeroconf, "_touch-able._tcp");
                            }
                        }
                    }
                }
            }
        }
    }

    //----------------------------------------------------------------------------

    private String hex(long l) {
        String v = Long.toHexString(l).toUpperCase();
        return ("000000000000000"+v).substring(v.length() - 1);
    }

    private ServerState getDaapInstance() {
        if (!sharing) {
            return null;
        } else {
            String dbid = hex(server.getUniqueID());
            ServerState state = new ServerState(context.getServerName(), null);
            state.props.put("txtvers", "1");
            state.props.put("Database ID", dbid);
            state.props.put("Machine ID", dbid);    // Doesn't have to be DBID, but works
            state.props.put("dmv", Integer.toString(DaapServer.MPRO_VERSION));
            state.props.put("Version", Integer.toString(DaapServer.APRO_VERSION));
            state.props.put("Media Kinds Shared", "1");       // Music only
            state.props.put("Machine Name", context.getServerName());
            state.props.put("Password", server.getPassword() != null ? "1" : "0");
//                state.props.put("OSsi", "0x1F5");
//                state.props.put("MID", "0x"+hex(server.getUniqueID()));
            return state;
        }
    }

    private ServerState getTouchableInstance(Player player) {
        if (remotemanager.hasRemotes(player)) {
            String dbid = hex(server.getUniqueID());
            String plid = hex(DaapServer.getUniqueID(player));
            ServerState state = new ServerState(plid, player);
            state.props.put("txtvers", "1");
            state.props.put("DbId", dbid);
            state.props.put("CtlN", player.getDisplayName()+" on "+context.getServerName());
            state.props.put("OSsi", "0xC9");
            state.props.put("DvSv", "2305");
            state.props.put("Ver", "131073");
            state.props.put("DvTy", "iTunes");
            return state;
        } else {
            return null;
        }
    }

    /**** Not used, although it seems to be advertised by iTunes alongside touch-able
    private ServerState getDacpInstance(Player player) {
        if (true) return null;
        DaapServer server = getControlServer(player);
        if (server == null) {
            return null;
        } else {
            String dbid = hex(server.getUniqueID());
            String plid = hex(server.getUniqueID());
            ServerState state = new ServerState(server, "iTunes_Ctrl_"+plid, player); // Arbitrary name
            state.props.put("txtvers", "1");
            state.props.put("DbId", dbid);
//                state.props.put("OSsi", "0x1F5");
//                state.props.put("Ver", "131075");
            return state;
        }
    }
    */

    private class ServerState {
        private final String serviceid;
        private final Player player;
        private ZeroConf zeroconf;
        private ZCService service;
        final Map<String,String> props;

        ServerState(String serviceid, Player player) {
            this.serviceid = serviceid;
            this.player = player;
            this.props = new LinkedHashMap<String,String>();
        }

        public String toString() {
            return player==null ? serviceid : player.getDisplayName();
        }

        public boolean equals(Object o) {
            ServerState s = (ServerState)o;
            return s==null ? false : serviceid.equals(s.serviceid) && props.toString().equals(s.props.toString());
        }

        public int hashCode() {
            return serviceid.hashCode() ^ props.toString().hashCode();
        }

        void register(ZeroConf zeroconf, String servicetype) {
            this.zeroconf = zeroconf;
            this.service = zeroconf.register(servicetype, serviceid, port, props);
        }

        void unregister() {
            zeroconf.unregister(service);
        }
    }

}
