package org.qtunes.db;

import java.util.*;

/**
 * This class lists the various Fields that can be part of a Track
 */
public final class Field<T> implements AbstractField<T> {

    /**
     * The value returned by {@link Field#getSearch}, this determines
     * whether the field should be indexed, searchable or neither when
     * stored in Lucene. Not currently used.
     */
    public enum Search {
        NONE, INDEX, SEARCH
    };
    private static final List<Field<?>> allreal = new ArrayList<Field<?>>();
    private static final Map<String,AbstractField<?>> all = new HashMap<String,AbstractField<?>>();

    public static final Field<String> Title = create("Title", String.class, Search.SEARCH);
    public static final Field<String> Artist = create("Artist", String.class, Search.SEARCH);
    public static final Field<String> Album = create("Album", String.class, Search.SEARCH);
    public static final Field<String> AlbumArtist = create("AlbumArtist", String.class, Search.SEARCH);
    public static final Field<String> Genre = create("Genre", String.class, Search.NONE);
    public static final Field<Integer> Track = create("Track", Integer.class, Search.INDEX);
    public static final Field<Integer> NumTracks = create("NumTracks", Integer.class, Search.NONE);
    public static final Field<Integer> Disc = create("Disc", Integer.class, Search.INDEX);
    public static final Field<Integer> NumDiscs = create("NumDiscs", Integer.class, Search.NONE);
    public static final Field<String> Composer = create("Composer", String.class, Search.SEARCH);
    public static final Field<Integer> Year = create("Year", Integer.class, Search.INDEX);
    public static final Field<Boolean> Compilation = create("Compilation", Boolean.class, Search.NONE);
    public static final Field<String> Comment = create("Comment", String.class, Search.SEARCH);
    public static final Field<String> CDDB = create("CDDB", String.class, Search.NONE);
    public static final Field<String> MCDI = create("MCDI", String.class, Search.NONE);
    public static final Field<String> Encoder = create("Encoder", String.class, Search.NONE);
    public static final Field<String> EncoderSettings = create("EncoderSettings", String.class, Search.NONE);
    public static final Field<String> Group = create("Group", String.class, Search.SEARCH);
    public static final Field<String> Language = create("Language", String.class, Search.NONE);
    public static final Field<String> Publisher = create("Publisher", String.class, Search.SEARCH);
    public static final Field<String> URL = create("URL", String.class, Search.NONE);
    public static final Field<Float> BPM = create("BPM", Float.class, Search.NONE);
    public static final Field<String> UserInfo = create("UserInfo", String.class, Search.SEARCH);
    public static final Field<String> ISRC = create("ISRC", String.class, Search.NONE);
    public static final Field<String> Media = create("Media", String.class, Search.NONE);
    public static final Field<String> Copyright = create("Copyright", String.class, Search.NONE);
    public static final Field<String> OriginalArtist = create("OriginalArtist", String.class, Search.SEARCH);
    public static final Field<String> OriginalAlbum = create("OriginalAlbum", String.class, Search.SEARCH);
    public static final Field<String> Conductor = create("Conductor", String.class, Search.SEARCH);
    public static final Field<String> Lyricist = create("Lyricist", String.class, Search.SEARCH);

    public static final Field<String> Type = create("Type", String.class, Search.NONE);
    public static final Field<String> Codec = create("Codec", String.class, Search.NONE);
    public static final Field<String> File = create("File", String.class, Search.INDEX);
    public static final Field<Integer> BitRate = create("BitRate", Integer.class, Search.NONE);  // Avg bitrate for VBR
    public static final Field<Integer> SampleRate = create("SampleRate", Integer.class, Search.NONE);
    public static final Field<Boolean> Mono = create("Mono", Boolean.class, Search.NONE);
    public static final Field<Boolean> VBR = create("VBR", Boolean.class, Search.NONE);
    public static final Field<Integer> Duration = create("Duration", Integer.class, Search.INDEX);  // In ms
    public static final Field<Integer> FileSize = create("FileSize", Integer.class, Search.INDEX);
    public static final Field<Integer> Added = create("Added", Integer.class, Search.INDEX);               // In seconds
    public static final Field<Integer> LastModified = create("LastModified", Integer.class, Search.INDEX); // In seconds
    public static final Field<Float> Gain = create("Gain", Float.class, Search.NONE);
    public static final Field<Boolean> Gapless = create("Gapless", Boolean.class, Search.NONE);

    public static final Field<Boolean> ID3v1  = create("ID3v1",  Boolean.class, Search.NONE);
    public static final Field<Boolean> ID3v22 = create("ID3v22", Boolean.class, Search.NONE);
    public static final Field<Boolean> ID3v23 = create("ID3v23", Boolean.class, Search.NONE);
    public static final Field<Boolean> ID3v24 = create("ID3v24", Boolean.class, Search.NONE);
    public static final Field<Boolean> Warning = create("Warning", Boolean.class, Search.NONE);

    public static final Field<String> ImageDescription = create("ImageDescription", String.class, Search.NONE);
    public static final Field<String> ImageMIMEType = create("ImageMIMEType", String.class, Search.NONE);
    public static final Field<Integer> ImageOffset = create("ImageOffset", Integer.class, Search.NONE);
    public static final Field<Integer> ImageLength = create("ImageLength", Integer.class, Search.NONE);
    public static final Field<Long> AudioByteRange = create("AudioByteRange", Long.class, Search.NONE);

    public static final Field[] ALLFIELDS;
    static {
        ALLFIELDS = (Field[])allreal.toArray(new Field[0]);
    }

    //-------------------------------------------------------------------------------

    private final Search searchtype;
    private final int index;
    private final String name;
    private final Class type;

    private Field(int index, String name, Class type, Search searchtype) {
        this.index = index;
        this.name = name;
        this.type = type;
        this.searchtype = searchtype;
    }

    private static <T> Field<T> create(String name, Class<T> type, Search searchtype) {
        Field<T> field = new Field<T>(allreal.size(), name, type, searchtype);
        register(field);
        allreal.add(field);
        return field;
    }

    protected static void register(AbstractField<?> field) {
        if (all.put(field.getName().toLowerCase(), field)!=null) {
            throw new IllegalStateException("Field "+field.getName()+" already exists");
        }
    }

    public static AbstractField getByName(String name) {
        return all.get(name.toLowerCase());
    }

    public int hashCode() {
        return index;
    }

    public String toString() {
        return name+"("+index+")";
    }

    public String getName() {
        return name;
    }

    public int getIndex() {
        return index;
    }

    public Class getType() {
        return type;
    }

    public Search getSearch() {
        return searchtype;
    }

}
