package org.qtunes.db.spi.simple;

import org.qtunes.core.*;
import org.qtunes.db.*;
import org.qtunes.ff.*;
import java.io.*;
import java.util.*;
import java.util.concurrent.locks.*;

/**
 * The Database manages the Tracks and Playlists. It stores
 * the Playlists as properties for this service.
 * </p><p>
 * The track database is done via a {@link Store}, and works
 * in two stages. First, a {@link MemoryStore} is updated with
 * the track information, then when {@link #commit} is called
 * this temporary store is merged with the main Store (probably
 * a {@link FileStore} at this stage. This is a flat-file which
 * is designed for very quick random access to track data.
 * </p><p>
 * This class is necessarily one of the more complex and it may
 * change over time. It must be thread safe (with java.util.concurrent).
 */
public class DatabaseImpl implements Database, Service {

    private ServiceContext context;
    private MemoryStore tempstore;              // The temporary store for updates
    private Store store;                        // The final store for permanent save
    private Map<Object,Track> inverse;          // A map from uniq->Track, generated from the DB
    private Matcher matcher;                    // Determines if a call to create() inserts or updates
    private int autocommit, revision, max;
    private Map<String,PlaylistVersion> playlists;
    private Map<Integer,Track> tracks;
    private Random random;

    private ReentrantReadWriteLock dblock = new ReentrantReadWriteLock();       // dblock, for tracks
    private ReentrantReadWriteLock pllock = new ReentrantReadWriteLock();       // pllock, for playlists

    private SmartPlaylist ALL;

    @SuppressWarnings("unchecked")
    public void startService(ServiceContext context) {
        this.context = context;
        this.random = new Random();
        this.playlists = new LinkedHashMap<String,PlaylistVersion>();
        this.store = "simple".equals(context.getProperty("storetype")) ? new FileStore() : new LuceneStore();
        this.matcher = new FilenameMatcher();
        try {
            store.setFile(context.getProperty("file"));
            try {
                autocommit = Integer.parseInt(context.getProperty("autocommit"));
            } catch (Exception e) {
                autocommit = 1000;
            }

            String[] propertynames = context.getPropertyNames();
            for (int i=0;i<propertynames.length;i++) {
                if (propertynames[i].startsWith("playlist.")) {
                    String name = propertynames[i].substring(9);
                    String value = context.getProperty(propertynames[i]);
                    if (value.charAt(0)=='[' && value.charAt(value.length()-1)==']') {
                        String[] s = value.substring(1, value.length()-1).split(",");
                        List<Track> tracks = new ArrayList<Track>(s.length);
                        for (int j=0;j<s.length;j++) {
                            Track track = getTrack(Integer.parseInt(s[j]));
                            if (track!=null) {
                                tracks.add(track);
                            }
                        }
                        ManualPlaylistImpl pl = new ManualPlaylistImpl(this, random.nextInt()&0x7FFFFFFF, tracks);
                        pl.setName(name);
                        playlists.put(name, new PlaylistVersion(pl));
                    } else {
                        SmartPlaylistImpl pl = new SmartPlaylistImpl(this, random.nextInt()&0x7FFFFFFF, value);
                        pl.setName(name);
                        playlists.put(name, new PlaylistVersion(pl));
                    }
                }
            }
        } catch (Exception e) {
            throw new IllegalStateException("Invalid properties", e);
        }
    }

    public void stopService(ServiceContext context) {
    }

    public ServiceContext getContext() {
        return context;
    }

    //--------------------------------------------------------------------------------

    /**
     * Create a new Track entry. The track will initially be created in the temporary
     * store - whether this method runs an "update" or an "insert" depends on the
     * {@link Matcher}, which will try to match the fields in the {@link FieldMap}
     * against the existing tracks. This may be done by Filename, by Inode or by 
     * checksum of the audio data (for example). The fields are then stored
     * and the Track returned.
     */
    public Track create(FieldMap map) {
        if (map.get(Field.File) == null) {
            throw new IllegalArgumentException("Can't add track with no File");
        }
        if (map.get(Field.Title) == null) {
            throw new IllegalArgumentException("Can't add track with no Title ("+map.get(Field.File)+")");
        }
        try {
            dblock.writeLock().lock();
            begin();
            if (inverse==null) {
                inverse = new HashMap<Object,Track>();
                Collection<Track> all = getTracks();
                for (Iterator<Track> i = all.iterator();i.hasNext();) {
                    Track track = i.next();
                    Object key = matcher.getKey(track);
                    inverse.put(key, track);
                }
            }
            Object key = matcher.getKey(map);
            Track oldtrack = (Track)inverse.get(key);
            int index = oldtrack==null ? ++max : oldtrack.getIndex();
            if (oldtrack==null && !map.containsKey(Field.Added)) {
                map.put(Field.Added, (int)(System.currentTimeMillis() / 1000));
            }
            for (Iterator<Map.Entry<Field<?>,Object>> i = map.entrySet().iterator();i.hasNext();) {
                Map.Entry<Field<?>,Object> e = i.next();
                tempstore.set(index, e.getKey(), e.getValue());
            }
            if (autocommit!=0 && tempstore.getTracks().cardinality() >= autocommit && store.isValid()) {
                commit();
                begin();
            }
            loadTracks();
            synchronized(tracks) {
                TrackImpl track = new TrackImpl(index, this);
                tracks.put(index, track);
                return track;
            }
        } finally {
            dblock.writeLock().unlock();
        }
    }

    private synchronized void loadTracks() {
        if (tracks == null) {
            try {
                dblock.readLock().lock();
                BitSet bs = (BitSet)store.getTracks().clone();
                tracks = new HashMap<Integer,Track>(bs.size());
                for (int i=bs.nextSetBit(0);i>=0;i=bs.nextSetBit(i+1)) {
                    tracks.put(i, new TrackImpl(i, this));
                    max = Math.max(max, i);
                }
            } finally {
                dblock.readLock().unlock();
            }
        }
    }

    public Collection<Track> getTracks() {
        loadTracks();
        synchronized(tracks) {
            return Collections.unmodifiableCollection(new HashSet<Track>(tracks.values()));
        }
    }

    public List<Track> getTracks(String sort, String filter) {
        loadTracks();
        try {
            dblock.readLock().lock();
            int[] ixlist = store.getTracks(sort, filter);
            List<Track> filteredtracks = new ArrayList<Track>(ixlist.length);
            for (int i=0;i<ixlist.length;i++) {
                Track track = tracks.get(Integer.valueOf(ixlist[i]));
                if (track != null) {
                    filteredtracks.add(track);
                }
            }
            return filteredtracks;
        } finally {
            dblock.readLock().unlock();
        }
    }

    public int getRevision() {
        return Math.max(revision, store.getRevision());
    }

    public int size() {
        loadTracks();
        synchronized(tracks) {
            return tracks.size();
        }
    }

    /**
     * Get the track with the specified track index
     */
    public Track getTrack(int tracknumber) {
        try {
            dblock.readLock().lock();
            getTracks();
            return tracks.get(tracknumber);
        } finally {
            dblock.readLock().unlock();
        }
    }

    Object getField(int track, String fieldname) {
        AbstractField field = Field.getByName(fieldname);
        if (field==null) {
            throw new IllegalArgumentException("Unknown field \""+fieldname+"\"");
        }
        if (field instanceof DerivedField) {
            return ((DerivedField)field).get(getTrack(track));
        } else {
            return getField(track, (Field)field);
        }
    }

    /**
     * Actually get the track field - called from Track.get() only
     */
    Object getField(int track, Field field) {
        Object val;
        if (tempstore!=null && tempstore.getTracks().get(track)) {
            val = tempstore.get(track, field);
        } else {
            val = store.get(track, field);
        }
        if (val == null && (field == Field.Title || field == Field.File)) {
            throw new Error("track="+track+" "+(field==Field.Title?"title is null":"file is null")+" ts="+(tempstore==null?null:tempstore.getTracks().get(track))+" s="+store.get(track,field));
        }
        return val;
    }

    /**
     * Get the set of fields set for this track
     */
    BitSet getFields(int track) {
        if (tempstore!=null && tempstore.getTracks().get(track)) {
            return tempstore.getFields(track);
        } else {
            return store.getFields(track);
        }
    }

    /**
     * Delete the specified track
     */
    void deleteTrack(int tracknumber) {
        try {
            dblock.writeLock().lock();
            begin();
            tempstore.delete(tracknumber);
            synchronized(tracks) {
                tracks.remove(tracknumber);
            }
        } finally {
            dblock.writeLock().unlock();
        }
    }

    public void lock(boolean hold) {
        if (hold) {
            dblock.readLock().lock();
        } else {
            dblock.readLock().unlock();
        }
    }

    public void commit() {
        if (store.isValid()) {
            new Thread("qTunes-db-commit") {
                public void run() {
                    try {
                        dblock.writeLock().lock();
                        if (tempstore!=null) {
                            if (tempstore.getRevision() != 0) {
                                int num = tempstore.getTracks().cardinality();
                                context.info("Committing "+num+" tracks");
                                store.merge(tempstore);
                                context.fireEvent("update", null);
                            }
                            tempstore = null;
                        }
                        savePlaylists(false, null);
                    } catch (RuntimeException e) {
                        e.printStackTrace();
                        throw e;
                    } catch (Exception e) {
                        e.printStackTrace();
                        throw new RuntimeException(e);
                    } finally {
                        dblock.writeLock().unlock();
                    }
                }
            }.start();
        }
    }

    /**
     * Abort the current pending transactions
     */
    public void abort() {
        tempstore = null;
    }

    private void begin() {
        if (tempstore==null) {
            tempstore = new MemoryStore();
        }
    }

    /**
     * Return true if a transaction has been begun with a call to {@link #begin})
     */
    public boolean hasPending() {
        return tempstore!=null;
    }

    public Playlist getPlaylist(String name) {
        if ("*".equals(name)) {
            if (ALL==null) {
                ALL = createSmartPlaylist("*");
            }
            return ALL;
        } else {
            try {
                pllock.readLock().lock();
                PlaylistVersion dbp = playlists.get(name);
                return dbp==null ? null : dbp.playlist;
            } finally {
                pllock.readLock().unlock();
            }
        }
    }

    public Collection<String> getPlaylists() {
        try {
            pllock.readLock().lock();
            List<String> names = new ArrayList<String>();
            names.addAll(playlists.keySet());
            names.remove(null);
            return names;
        } finally {
            pllock.readLock().unlock();
        }
    }

    public void putPlaylist(String name, Playlist playlist) {
        if (name == null || name.equals("*")) {
            throw new IllegalArgumentException("Invalid playlist name "+name);
        }
        if (playlist != null && !(playlist instanceof ManualPlaylistImpl || playlist instanceof SmartPlaylistImpl)) {
            throw new IllegalArgumentException("Can't save this type of playlist: "+playlist.getClass().getName());
        }
        try {
            pllock.writeLock().lock();
            boolean statechanged = false;
            Collection<String> playlistnames = new LinkedHashSet<String>();
            if (playlist != null) {
                String oldname = playlist.getName();
                PlaylistVersion pv = playlists.get(name);
                if (pv != null && pv.playlist != playlist) {
                    throw new IllegalArgumentException("Playlist \""+name+"\" already exists");
                }
                if (oldname == null || !name.equals(oldname)) {
                    statechanged = true;
                    pv = playlists.remove(oldname);
                    if (pv == null) {
                        pv = new PlaylistVersion(playlist);
                    }
                    playlists.put(name, pv);
                    playlist.setName(name);
                    if (oldname != null) {
                        playlistnames.add(oldname);
                    }
                    playlistnames.add(name);
                }
            } else if (playlists.get(name) != null) {
                playlists.put(name, null);
                statechanged = true;
                playlistnames.add(name);
            }
            savePlaylists(statechanged, playlistnames);
        } finally {
            pllock.writeLock().unlock();
        }
    }

    public void updateTrack(int tracknumber, FieldMap changes) throws IOException {
        try {
            dblock.writeLock().lock();
            begin();
            Track track = getTrack(tracknumber);
            TrackReader reader = TrackReader.getTrackReader(context, (String)track.get(Field.Type));
            if (!(reader instanceof TrackWriter)) {
                throw new IllegalStateException("Tracks of type \""+track.get(Field.Type)+"\" cannot be updated");
            }
            context.info("Updating "+track.getFile()+" with "+changes);
            ((TrackWriter)reader).write(track, changes);
            for (int i=0;i<Field.ALLFIELDS.length;i++) {
                Field<?> field = Field.ALLFIELDS[i];
                Object value = changes.get(field);
                if (value==null) {
                    value = store.get(tracknumber, field);
                }
                if (value!=null) {
                    tempstore.set(tracknumber, field, value);
                }
            }
            context.fireEvent("update", null);
            context.fireEvent("trackEdit", new Object[] { "track", track });
        } finally {
            dblock.writeLock().unlock();
        }
    }

    private void savePlaylists(boolean statechanged, Collection<String> names) {
        try {
            pllock.writeLock().lock();

            // Save state?
            if (names == null) {
                names = new LinkedHashSet<String>();
            }
            for (Iterator<Map.Entry<String,PlaylistVersion>> i = playlists.entrySet().iterator();i.hasNext();) {
                Map.Entry<String,PlaylistVersion> e = i.next();
                if (e.getValue()==null) {
                    statechanged = true;
                    names.add(e.getKey());
                    i.remove();
                } else if (e.getValue().revision != e.getValue().playlist.getRevision()) {
                    e.getValue().revision = e.getValue().playlist.getRevision();
                    statechanged = true;
                    names.add(e.getKey());
                }
            }
            if (statechanged) {
                for (Iterator<String> i=names.iterator();i.hasNext();) {
                    String name = i.next();
                    PlaylistVersion pv = playlists.get(name);
                    if (pv==null) {
                        context.putProperty("playlist."+name, null);
                    } else if (pv.playlist instanceof SmartPlaylist) {
                        context.putProperty("playlist."+name, ((SmartPlaylist)pv.playlist).getQuery());
                    } else {
                        List<Track> pltracks = pv.playlist.getTracks();
                        StringBuffer sb = new StringBuffer();
                        for (int j=0;j<pltracks.size();j++) {
                            sb.append(j==0?'[':',');
                            sb.append(pltracks.get(j).getIndex());
                        }
                        sb.append(']');
                        context.putProperty("playlist."+name, sb.toString());
                    }
                }
                revision++;
                context.fireEvent("update", null);
                context.fireEvent("playlistEdit", new Object[] { "names", names });
            }
        } finally {
            pllock.writeLock().unlock();
        }
    }

    private static final class PlaylistVersion {
        Playlist playlist;
        int revision;
        PlaylistVersion(Playlist playlist) {
            this.playlist = playlist;
            this.revision = 0;
        }
        public String toString() {
            return "[v="+revision+" p="+playlist+"]";
        }
    }

    public ManualPlaylist createManualPlaylist() {
        return new ManualPlaylistImpl(this, random.nextInt()&0x7FFFFFFF, null);
    }

    public SmartPlaylist createSmartPlaylist(String query) {
        return new SmartPlaylistImpl(this, random.nextInt()&0x7FFFFFFF, query);
    }

    public Map<String,Object> reportState() {
        Map<String,Object> map = new LinkedHashMap<String,Object>();
        map.put("size", size());
        map.put("revision", getRevision());
        Map<String,Map<String,Object>> playlists = new HashMap<String,Map<String,Object>>();
        for (Iterator<String> i = getPlaylists().iterator();i.hasNext();) {
            String name = i.next();
            Playlist pl = getPlaylist(name);
            int id = pl.getUniqueID();
            Map<String,Object> props = new LinkedHashMap<String,Object>();
            props.put("name", name);
            props.put("size", pl.size());
            props.put("revision", pl.getRevision());
            props.put("smart", pl instanceof SmartPlaylist);
            if (pl instanceof SmartPlaylist) {
                props.put("query", ((SmartPlaylist)pl).getQuery());
            }
            playlists.put(Integer.toString(id), props);
        }
        map.put("playlists", playlists);
        return map;
    }

}
