package org.qtunes.main;

import org.qtunes.core.*;
import org.qtunes.auth.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.util.concurrent.*;
import org.apache.log4j.*;

/**
 * The main class. The qTunes server does very little on its own, but mainly
 * handle the loading/saving of the configuration and state, and the starting
 * and stopping of {@link Service} objects.
 */
public class QTunes {

    private String name;
    private Date builddate;
    private PropertyManager properties;
    public List<ServiceContextImpl> contexts;
    private List<Listener> listeners;
    private final Logger log = Logger.getLogger(QTunes.class);

    /**
     * Create a new QTunes server
     */
    public QTunes() {
        contexts = new CopyOnWriteArrayList<ServiceContextImpl>();
        listeners = new CopyOnWriteArrayList<Listener>();
        properties = new PropertyManager(this);

        BufferedReader r = null;
        try {
            String classname = "/" + getClass().getName().replace('.', '/')+".class";
            java.net.URL url = getClass().getResource(classname);
            url = new java.net.URL(url.toString().replaceAll(classname, "/META-INF/MANIFEST.MF"));
            r = new BufferedReader(new InputStreamReader(url.openStream(), "ISO-8859-1"));
            String s;
            while ((s=r.readLine())!=null) {
                if (s.startsWith("Build-Date:")) {
                    s = s.substring(11).trim();
                    builddate = new SimpleDateFormat("yyyyMMdd'T'HHmmss").parse(s);
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            if (r != null) {
                try { r.close(); } catch (IOException e) {}
            }
        }
    }

    /**
     * Return the name and version of the Software
     */
    public String getSoftwareName() {
        Package pack = getClass().getPackage();
        String title = pack.getImplementationTitle();
        String version = pack.getImplementationVersion();
        if (title == null || title.length()==0) {
            title = "qTunes";
        }
        if (version==null || version.length()==0) {
            version = "?";
        }
        return title+"/"+version+" (built on "+(new SimpleDateFormat("yyyyMMdd HH:mm:ss").format(builddate))+")";
    }

    /**
     * Return the name of the Server, as specified in the configuration
     */
    @ControlMethod (
        params = {},
        help = "Return the name of the Server, which should uniquely identify this install on the network"
    )
    public String getServerName() {
        return name;
    }

    public Date getServerBuildDate() {
        return builddate;
    }

    public Iterator<ServiceContextImpl> getServices() {
        return contexts.iterator();
    }

    public Iterator<ServiceContextImpl> getActiveServices() {
        List<ServiceContextImpl> copy = new ArrayList<ServiceContextImpl>(contexts);
        for (Iterator<ServiceContextImpl> i = copy.iterator();i.hasNext();) {
            if (!i.next().isActive()) {
                i.remove();
            }
        }
        return copy.iterator();
    }

    /**
     * Deregister all Services and quit the Server
     */
    public synchronized void quit() {
        for (int i=contexts.size()-1;i>=0;i--) {
            contexts.remove(i).stop();
        }
        notifyAll();
    }

    /**
     * Load the configuration from the specified file
     */
    public synchronized void loadConfiguration(File file) throws IOException {
        for (int i=contexts.size()-1;i>=0;i--) {
            contexts.get(i).stop();
        }
        contexts.clear();

        properties.parse(file);
        String value;
        if ((value=(String)properties.get("name"))!=null) {
            this.name = value;
        }
        if ((value=(String)properties.get("log4j.xml"))!=null) {
            // Don't import this so we can use log4j mini
            try {
                Class.forName("org.apache.log4j.xml.DOMConfigurator").getMethod("configure", String.class).invoke(null, value);
            } catch (Throwable e) {
                throw new RuntimeException(e);
            }
        } else if ((value=(String)properties.get("log4j.properties"))!=null) {
            PropertyConfigurator.configure(value);
        }
        info("Initialized "+getSoftwareName());
        info("Loaded configuration from \""+file.getAbsolutePath()+"\"");
        for (Iterator<ServiceContextImpl> i = contexts.iterator();i.hasNext();) {
            ServiceContextImpl sc = i.next();
            sc.start();
        }
        notifyAll();
    }

    public void addServiceContext(ServiceContextImpl context) {
        contexts.add(context);
    }

    public void removeServiceContext(ServiceContextImpl context) {
        contexts.remove(context);
        properties.remove(context.getProperty("name"));
    }

    //------------------------------------------------------------------------------------------

    public void addListener(Listener listener) {
        listeners.add(listener);
    }

    public void removeListener(Listener listener) {
        listeners.remove(listener);
    }

    public void fireEvent(String name, Map<?,?> eventprops) {
        Map<Object,Object> m = new HashMap<Object,Object>(eventprops);
        Service service = (Service)m.get("service");
        m.put("service", service.getContext().getProperty("name"));
        String message = "-> fire "+name+" "+m;
        service.getContext().debug(message);
        debug(message);
        for (Iterator i = listeners.iterator();i.hasNext();) {
            Listener listener = (Listener)i.next();
            listener.handleEvent(name, eventprops);
        }
    }

    void debug(String message) {
        log.debug(message);
    }

    void info(String message) {
        log.info(message);
    }

    void warn(String message, Throwable e) {
        log.warn(message, e);
    }

    void error(String message, Throwable e) {
        log.error(message, e);
    }

    public static void main(String[] args) throws Exception {
        if (args.length == 0) {
            System.out.println("Usage: org.qtunes.QTunes [configfile]\n\n");
            System.exit(0);
        }
        GUI gui = null;
        File configfile = null;
        if (args[0].equals("--macgui")) {
            gui = new GUI();
            configfile = gui.chooseConfigFile();
            if (configfile == null) {
                System.exit(0);
            }
        } else {
            configfile = new File(args[0]);
        }

        final QTunes qtunes = new QTunes();
        qtunes.loadConfiguration(configfile);
        do {
            synchronized(qtunes) {
                qtunes.wait();
            }
        } while (!qtunes.getActiveServices().hasNext());
    }

}
