package org.qtunes.main;

import org.qtunes.core.*;
import org.qtunes.ff.*;
import org.qtunes.db.Database;
import java.io.IOException;
import java.io.File;
import java.lang.reflect.Array;
import java.util.*;
import org.apache.log4j.Logger;

public class ServiceContextImpl implements ServiceContext {

    private QTunes qtunes;
    private Service service;
    private Logger log;
    private PropertyManager props;
    private volatile boolean active;
    private final String servicename;

    public ServiceContextImpl(QTunes qtunes, Service service, String servicename, PropertyManager props) {
        this.qtunes = qtunes;
        this.service = service;
        this.log = Logger.getLogger(service.getClass());
        this.props = props;
        this.servicename = servicename;
    }

    public Service getService() {
        return service;
    }

    public String getServiceName() {
        return servicename;
    }

    public void debug(String message) {
        log.debug(message);
    }

    public void debug(String message, Throwable e) {
        log.debug(message, e);
    }

    public void info(String message) {
        log.info(message);
    }

    public void warn(String message, Throwable e) {
        log.warn(message, e);
    }

    public void error(String message, Throwable e) {
        log.error(message, e);
    }

    public String getServerName() {
        return qtunes.getServerName();
    }

    public String getSoftwareName() {
        return qtunes.getSoftwareName();
    }

    public Date getServerBuildDate() {
        return qtunes.getServerBuildDate();
    }

    /**
     * Add a {@link Listener} to the Server, it will be notified of events
     */
    public void addListener(Listener listener) {
        qtunes.addListener(listener);
    }

    /**
     * Remove a {@link Listener} to the Server, it will be notified of events
     */
    public void removeListener(Listener listener) {
        qtunes.removeListener(listener);
    }

    /**
     * Fire an event - the Listeners registered with the server will be notified.
     * Properties can be specified easily with the second parameter, which is a sequence
     * of [key, value] entries - eg
     * <pre>
     *  qtunes.fireEvent("stateChanged", new Object[] { "track", track, "user", user });
     * </pre>
     * @param o an array of [key, value]
     */
    public void fireEvent(String topic, Object[] o) {
        Map<Object,Object> eventprops = new HashMap<Object,Object>();
        if (o!=null) {
            for (int i=0;i<o.length;i+=2) {
                eventprops.put(o[i], o[i+1]);
            }
        }
        if (!eventprops.containsKey("session")) {
            Session session = Session.getCurrentSession();
            if (session!=Session.NONE) {
                eventprops.put("session", session);
            }
        }
        if (!eventprops.containsKey("service")) {
            eventprops.put("service", service);
        }
        eventprops.put("when", new Long(System.currentTimeMillis()));
        qtunes.fireEvent(topic, eventprops);
    }

    public String[] getPropertyNames() {
        return props.getNames(getServiceName()).toArray(new String[0]);
    }

    public String getGlobalProperty(String key) {
        return props.get(null, key);
    }

    public String getProperty(String key) {
        return props.get(getServiceName(), key);
    }

    public void putProperty(String key, String value) {
        props.put(getServiceName(), key, value);
    }

    public <E extends Service> E getService(Class<E> type) {
        return getService(type, null);
    }

    @SuppressWarnings("unchecked")
    public <E extends Service> E getService(Class<E> type, String filter) {
        ServiceMatcher matcher = filter==null ? null : new ServiceMatcher(filter);
        for (Iterator<ServiceContextImpl> i = qtunes.getServices();i.hasNext();) {
            Service service = i.next().getService();
            if ((type==null || type.isAssignableFrom(service.getClass())) && (matcher==null || matcher.matches(service))) {
                return (E)service;
            }
        }
        for (Iterator<ServiceContextImpl> i = qtunes.getServices();i.hasNext();) {
            Service t = i.next().getService();
            if (t instanceof ServiceFactory) {
                ServiceFactory factory = (ServiceFactory)t;
                Map<String,String> properties = matcher==null ? new HashMap<String,String>() : matcher.getProperties();
                Service service = factory.createService(type, properties);
                if (service!=null) {
                    Class<?>[] classes = factory.getServiceClasses().toArray(new Class[0]);
                    addService(classes, service, properties, false).start();
                    return (E)service;
                }
            }
        }
        return null;
    }

    @SuppressWarnings("unchecked")
    public <E extends Service> E[] getServices(Class<E> type, String filter) {
        ServiceMatcher matcher = filter==null ? null : new ServiceMatcher(filter);
        List<Object> all = new ArrayList<Object>();
        for (Iterator<ServiceContextImpl> i = qtunes.getServices();i.hasNext();) {
            Service service = i.next().getService();
            if ((type==null || type.isAssignableFrom(service.getClass())) && (matcher==null || matcher.matches(service))) {
                all.add(service);
            }
        }
        return all.toArray((E[])Array.newInstance(type, all.size()));
    }

    public void start() {
        info("Starting \""+getServiceName()+"\"");
        service.startService(this);
        active = true;
        fireEvent("started", null);
    }

    public void stop() {
        info("Stopping \""+getServiceName()+"\"");
        active = false;
        service.stopService(this);
        fireEvent("stopped", null);
        if (!props.isPermanent(getServiceName())) {
            props.remove(getServiceName());
        }
    }

    public boolean isActive() {
        return active;
    }

    public ServiceContext addService(Class<?>[] classes, Service service, Map<String,String> properties, boolean permanent) {
        for (int i=0;i<classes.length;i++) {
            if (!classes[i].isAssignableFrom(service.getClass())) {
                 throw new IllegalArgumentException("Service does not implement "+classes[i].getName());
            }
        }
        String servicename = properties.remove("name");
        if (servicename == null) {
            throw new IllegalStateException("No \"name\" property");
        }
        props.create(servicename, service.getClass(), properties, permanent);
        ServiceContextImpl impl = new ServiceContextImpl(qtunes, service, servicename, props);
        qtunes.addServiceContext(impl);
        return impl;
    }

    public void removeService() {
        qtunes.removeServiceContext(this);
    }

    public String[] getServiceNames() {
        List<String> names = new ArrayList<String>();
        for (Iterator<ServiceContextImpl> i = qtunes.getServices();i.hasNext();) {
            Service service = i.next().getService();
            if (service.getContext() != null) {
                names.add(service.getContext().getServiceName());
            }
        }
        return (String[])names.toArray(new String[0]);
    }

    @ControlMethod (
        params = {},
        auth = "admin",
        help = "Deregister all Services and quit the server"
    )
    public void quit() {
        qtunes.quit();
    }

}
