package org.qtunes.player.spi;

import org.qtunes.core.*;
import org.qtunes.player.*;
import org.qtunes.speaker.*;
import org.qtunes.db.*;
import org.qtunes.auth.*;
import java.util.*;
import java.net.URL;
import java.io.*;
import javax.sound.sampled.*;

/**
 * A Player plays tracks to one or more speakers (technically a {@link SourceDataLine}
 * which may be a local speaker, an Airtunes speaker or some other type (eg RTSP, uPnP etc.)
 * Each player maintains it's own playlist and tracks may be cued, reordered, skipped
 * and so on as you'd expect.
 */
public class PlayerImpl implements Player, Runnable, Listener {

    private static final int BUFSIZE = 4096;
    private ServiceContext context;
    private Database db;
    private int revision;
    private PlayerPlaylist queue;
    private volatile boolean paused, cancelled, trackinterrupted;
    private String mixername, mixervendor;
    private volatile TrackSession tracksession;
    private MultiLine multiline;
    private Thread thread;
    private volatile int volume, repeat;
    private List<String> pendingspeakernames;

    public PlayerImpl() {
        queue = new PlayerPlaylist(this);
        multiline = new MultiLine(this);
    }

    /**
     * Called when the player has changed - track seek, skipped, paused or unpaused
     */
    private void notifyPlayChanged() {
        revision++;
        context.fireEvent("playChange", new Object[] { });
    }

    /**
     * Called when the playlist has changed - tracks added or moved
     */
    synchronized void notifyListChanged(boolean startplaying) {
        if (queue.indexOf(tracksession) < 0) {
            if (startplaying && paused) {
                paused = false;
            }
            tracksession = startplaying ? queue.getTrackSession(0) : null;
            trackinterrupted = true;
        }
        notifyAll();
        revision++;
        context.fireEvent("playlistChange", new Object[] { });
    }

    /**
     * Called when a non-play event is changed (shuffle, volume, repeat etc.)
     */
    private void notifyStateChanged() {
        revision++;
        context.fireEvent("stateChange", new Object[] { });
    }


    /**
     * The main play thread. Should be bulletproof
     */
    public void run() {
        byte[] buf = new byte[BUFSIZE];
        AudioInputStream in = null;
        TrackSession lastplaying = null, nowplaying = null;
        int streamoffset = 0;

        context.fireEvent("started", null);
        boolean autoadvance = false;
        while (!cancelled) {
            synchronized(this) {
                trackinterrupted = false;
                lastplaying = nowplaying;
                tracksession = nowplaying = queue.getNextTrackSession(tracksession, autoadvance ? 1 : 0, autoadvance, repeat);
            }
            if (nowplaying == null) {
//                context.debug("Player wakes: now==null last"+(lastplaying==null?"==null":"!=null"));
                if (lastplaying != null) {
                    if (autoadvance) {
                        multiline.drain();
                    } else {
                        multiline.flush();
                    }
                    multiline.close();
                    lastplaying = null;
                }
                autoadvance = false;
                try {
                    synchronized(this) {
                        wait();
                    }
                } catch (InterruptedException e) {}
            } else {
                autoadvance = false;
                try {
                    in = nowplaying.track.getAudioInputStream();
                    streamoffset = 0;
                    nowplaying.setFormat(in.getFormat());
                    context.fireEvent("startPlaying", new Object[] { "track", nowplaying.track });
                    setVolume(getVolume());     // Adjust for per-track gain
//                    context.debug("Player wakes: now!=null track="+nowplaying.track.getIndex()+" last"+(lastplaying==null?"==null":lastplaying.formatMatches(nowplaying)?"==diff":"==same"));

                    if (lastplaying!=null && !lastplaying.formatMatches(nowplaying)) {
                        multiline.drain();
                        multiline.close();
                        lastplaying = null;
                    }
                    if (lastplaying == null) {
                        AudioFormat fmt = nowplaying.getFormat();
                        multiline.open(fmt.getSampleRate(), fmt.getSampleSizeInBits(), fmt.getChannels(), fmt.getEncoding()==AudioFormat.Encoding.PCM_SIGNED, fmt.isBigEndian());
                        setVolume(getVolume());
                    }

                    while (!trackinterrupted && !cancelled) {
                        boolean bumped = false;
                        if (in!=null && streamoffset > nowplaying.tell()) {
                            in.close();
                            in = null;
                        }
                        if (in==null) {
                            in = nowplaying.track.getAudioInputStream();
                            streamoffset = 0;
                            bumped = true;
                        }
                        int n;
                        while (nowplaying.tell() > streamoffset && (n=in.read(buf, 0, Math.min(nowplaying.tell() - streamoffset, buf.length)))>=0) {
                            bumped = true;
                            streamoffset += n;
                        }
                        if (paused) {
//                            context.debug("Player paused");
                            multiline.stop();
                            notifyPlayChanged();
                            try {
                                synchronized(this) {
                                    wait();
                                }
                            } catch (InterruptedException e) {}
                            bumped = true;
                        }
                        if (bumped) {
//                            context.debug("Player bumped");
                            notifyPlayChanged();
                        }
                        if (!paused) {
                            if (!multiline.hasSpeakers()) {
                                trackinterrupted = true;
                                context.fireEvent("error", new Object[] { "exception", new IllegalStateException("No Speakers for \""+getDisplayName()+"\"") });
                            } else if ((n = in.read(buf, 0, buf.length)) >= 0) {
                                streamoffset += n;
                                nowplaying.advance(n);
                                multiline.write(buf, 0, n);
                                nowplaying.session.bumpTraffic(n);
                            } else {
//                                context.debug("Player ends track");
                                context.fireEvent("stopPlaying", new Object[] { "track", nowplaying.track });
                                nowplaying.session.bumpTracksCompleted();
                                trackinterrupted = autoadvance = true;
                            }
                        }
                    }
//                    context.debug("Player ends loop: int="+trackinterrupted+" auto="+autoadvance+" cancelled="+cancelled);
                    if (trackinterrupted) {
                        if (!autoadvance) {
                            multiline.flush();
                        }
                        notifyPlayChanged();
                    }
                } catch (Exception e) {
                    context.warn("Play failed", e);
                    if (in!=null) {
                        try { in.close(); } catch (Exception e2) {} 
                        in = null;
                    }
                    if (multiline.isOpen()) {
                        try {
                            multiline.flush();
                            multiline.close();
                        } catch (Exception e2) {}
                    }
                    lastplaying = null;
                    autoadvance = true;
                }
            }
        }
//        context.debug("Player dies");
    }

    synchronized void interrupt() {
        trackinterrupted = true;
        notifyAll();
    }

    //----------------------------------------------------------------------------------

    public synchronized void playNext() {
        tracksession = queue.getNextTrackSession(tracksession, 1, false, repeat);
        interrupt();
    }

    public synchronized void playPrevious() {
        if (getCurrentTrackIndex()==0) {
            seek(0);
        } else {
            tracksession = queue.getNextTrackSession(tracksession, -1, false, repeat);
            interrupt();
        }
    }

    public synchronized void setCurrentTrackIndex(int ix) {
        tracksession = queue.getTrackSession(ix);
        interrupt();
    }

    public synchronized void stop() {
        tracksession = null;
        interrupt();
    }

    public synchronized void seek(int ms) {
        if (tracksession != null) {
            tracksession.seekMS(ms);
        }
    }

    public synchronized void setPaused(boolean paused) {
        if (this.paused != paused && tracksession != null) {
            this.paused = paused;
            notifyAll();
        }
    }

    public boolean isPaused() {
        return paused;
    }

    public int getTrackPosition() {
        return Math.max(0, tracksession==null ? 0 : tracksession.tellMS() - multiline.getDelay());
    }

    public synchronized long getRemaining() {
        long time = 0;
        if (tracksession != null) {
            time = (tracksession.track.get(Field.Duration)) - getTrackPosition();
        }
        TrackSession ts = tracksession;
        while ((ts=queue.getNextTrackSession(ts, 1, false, 0)) != null) {
            time += ts.track.get(Field.Duration);
        }
        return time;
    }

    public synchronized void setShuffle(boolean shuffle) {
        if (shuffle != queue.isShuffle()) {
            queue.setShuffle(shuffle);
            notifyStateChanged();
        }
    }

    public void setRepeat(int repeat) {
        if (repeat != this.repeat) {
            this.repeat = repeat;
            notifyStateChanged();
        }
    }

    public int getRepeat() {
        return repeat;
    }

    public boolean getShuffle() {
        return queue.isShuffle();
    }

    public synchronized int getCurrentTrackIndex() {
        return queue.indexOf(tracksession);
    }

    /**
     * Return the currently playing Track, or null if no track is playing
     */
    public Track getCurrentTrack() {
        return tracksession == null ? null : tracksession.track;
    }

    public void setVolume(int volume) {
        volume = Math.max(0, Math.min(100, volume));
        float trackgain = 0;
        if (tracksession != null) {
            trackgain = tracksession.track.get(Field.Gain);
            if (trackgain != trackgain) {
                trackgain = 0;
            }
        }
        multiline.setVolume(volume, trackgain);
        if (volume != this.volume) {
            this.volume = volume;
            context.putProperty("volume", Integer.toString(volume));
            notifyStateChanged();
        }
    }

    public int getVolume() {
        return volume;
    }

    public int getRevision() {
        return revision;
    }

    public String getDisplayName() {
        return context.getProperty("display");
    }

    /**
     * Return the Playlist
     */
    public ManualPlaylist getPlaylist() {
        return queue;
    }

    //-------------------------------------------------------------------------------

    public Collection<Speaker> getAvailableSpeakers() {
        Collection<Speaker> speakers = new ArrayList<Speaker>();
        Speaker[] all = context.getServices(Speaker.class, null);
        for (int i=0;i<all.length;i++) {
            Speaker speaker = all[i];
            if (speaker.getPlayer() == null || speaker.getPlayer() == this) {
                speakers.add(speaker);
            }
        }
        return speakers;
    }

    public void addSpeaker(Speaker speaker) {
        Collection<Speaker> speakers = multiline.getSpeakers();
        if (speakers.add(speaker)) {
            multiline.setSpeakers(speakers);
            updateSpeakerProperties();
        }
    }

    public void removeSpeaker(Speaker speaker) {
        Collection<Speaker> speakers = multiline.getSpeakers();
        if (speakers.remove(speaker)) {
            multiline.setSpeakers(speakers);
            updateSpeakerProperties();
        }
    }

    public Collection<Speaker> getSpeakers() {
        return multiline.getSpeakers();
    }

    private void updateSpeakerProperties() {
        int j = 1;
        Collection<Speaker> speakers = getSpeakers();
        for (Iterator<Speaker> i = speakers.iterator();i.hasNext();) {
            Speaker speaker = i.next();
            getContext().putProperty("speaker."+(j++), speaker.getContext().getServiceName());
        }
        while (getContext().getProperty("speaker."+j)!=null) {
            getContext().putProperty("speaker."+(j++), null);
        }
        notifyStateChanged();
    }

    //-----------------------------------------------------------------------------

    public void startService(final ServiceContext context) {
        this.context = context;
        this.cancelled = false;
        this.tracksession = null;
        this.pendingspeakernames = new ArrayList<String>();

        db = context.getService(Database.class);
        String volstate = context.getProperty("volume");
        try {
            volume = Integer.parseInt(volstate);
        } catch (Exception e) {
            volume = 50;
        }
        if (context.getProperty("display")==null) {
            throw new IllegalStateException("No \"display\" property");
        }

        String[] propnames = context.getPropertyNames();
        for (int i=0;i<propnames.length;i++) {
            if (propnames[i].startsWith("speaker.")) {
                pendingspeakernames.add(context.getProperty(propnames[i]));
            }
        }

        thread = new Thread(this, "qTunes-"+context.getServiceName());
        thread.start();

        for (Iterator<String> i = pendingspeakernames.iterator();i.hasNext();) {
            String speakername = i.next();
            Speaker speaker = context.getService(Speaker.class, "name=\""+speakername+"\"");
            if (speaker != null) {
                i.remove();
                try {
                    speaker.setPlayer(PlayerImpl.this);
                } catch (Exception e) {
                    context.warn("Can't set speaker \""+speakername+"\"", e);
                }
            } else {
                context.warn("Can't find speaker \""+speakername+"\"", null);
            }
        }
        // Add speakers we expected on load but weren't started until after
        // we've started.
        context.addListener(this);
    }

    public void stopService(ServiceContext context) {
        cancelled = trackinterrupted = true;
        tracksession = null;
        thread.interrupt();
        try {
            thread.join();
        } catch (InterruptedException e) { }
        multiline = null;
        context.removeListener(this);
    }

    public ServiceContext getContext() {
        return context;
    }

    public Map<String,Object> reportState() {
        Map<String,Object> map = new LinkedHashMap<String,Object>();
        map.put("name", getDisplayName());
        map.put("revision", getRevision());
        map.put("volume", getVolume());
        map.put("shuffle", getShuffle());
        map.put("paused", isPaused());
        map.put("repeat", getRepeat());
        Collection<Speaker> speakers = getSpeakers();
        List<String> speakerlist = new ArrayList<String>();
        for (Iterator<Speaker> i = speakers.iterator();i.hasNext();) {
            Speaker speaker = i.next();
            speakerlist.add(speaker.getContext().getProperty("name"));
        }
        map.put("speakers", speakerlist);

        Track track = getCurrentTrackIndex() >= 0 ? getCurrentTrack() : null;
        map.put("playlistindex", getCurrentTrackIndex());
        map.put("playlistlength", queue.size());
        if (track != null) {
            map.put("playoffset", getTrackPosition());
            map.put("totalplayremaining", getRemaining());
            map.put("np_index", track.getIndex());
            map.put("np_duration", Integer.valueOf(track.get(Field.Duration)));
            map.put("np_title", track.get(Field.Title));
            map.put("np_artist", track.get(Field.Artist));
            map.put("np_album", track.get(Field.Album));
            map.put("np_albumartist", track.get(Field.AlbumArtist));
            map.put("np_track", track.get(Field.Track));
            map.put("np_numtracks", track.get(Field.NumTracks));
            map.put("np_disc", track.get(Field.Disc));
            map.put("np_numdiscs", track.get(Field.NumDiscs));
            map.put("np_year", track.get(Field.Year));
            map.put("np_genre", track.get(Field.Genre));
        }
//        context.debug("PLAYER="+map);
        return map;
    }

    public void handleEvent(String name, java.util.Map<?,?> properties) {
        Service service = null;
        if ((service=(Service)properties.get("service")) instanceof Speaker) {
            Speaker speaker = (Speaker)service;
            if (name.equals("started")) {
                String speakername = speaker.getContext().getServiceName();
                if (speaker.getPlayer() == null && pendingspeakernames.remove(speakername)) {
                    speaker.setPlayer(this);
                }
            } else if (speaker.getPlayer() == PlayerImpl.this && name.equals("speakerError")) {
                stop();
                context.fireEvent("error", new Object[] { "exception", properties.get("exception"), "speaker", speaker });
            }
        }
    }

}
