package org.qtunes.replaygain;

import org.qtunes.core.util.FileUtils;
import org.qtunes.ff.TrackReader;
import java.io.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.zip.*;
import javax.sound.sampled.*;
import javax.xml.stream.*;

/**
 * A class which manages the ReplayGain normalization on a number of files.
 * It stores the {@link GainAnalysis} to a database, queues new analyses
 * to ensure they're not all run at once or repeated if we've already analysed
 * a file.
 */
public class Normalizer {

    private PrintStream out = System.out;
    private boolean debug;
    private File dbfile;
    private long lastsave, lastmodified, autosave;
    private ThreadPoolExecutor executor;
    private ConcurrentMap<String,GainAnalysis> all;
    private ConcurrentMap<String,Object> pending;
    private static final ThreadLocal<ReplayGainAnalyzer> analyzer = new ThreadLocal<ReplayGainAnalyzer>() {
        protected ReplayGainAnalyzer initialValue() {
            return new ReplayGainAnalyzer();
        }
    };

    public Normalizer() {
        executor = (ThreadPoolExecutor)Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
        pending = new ConcurrentHashMap<String,Object>();
        autosave = 30000;
    }

    /**
     * Set the output stream for info/debug
     * @param out the PrintStream to write debug to
     * @param debug whether to write debug or not
     */
    public void setDebug(PrintStream out, boolean debug) {
        this.out = out;
        this.debug = debug;
    }

    /**
     * Save the Database of analyses to the database file set in {@link #setDatabase}
     */
    public synchronized void saveDatabase() throws IOException {
        if (dbfile != null && lastsave != lastmodified) {
            lastsave = lastmodified = System.currentTimeMillis();
            File tempfile = File.createTempFile(dbfile.getName(), null, dbfile.getAbsoluteFile().getParentFile());
            long ms = System.currentTimeMillis();
            OutputStreamWriter writer = new OutputStreamWriter(new BufferedOutputStream(new FileOutputStream(tempfile)), "UTF-8");
            writer.write("<?xml version=\"1.0\"?>\n<normalization>\n");
            Set<String> keys = new TreeSet<String>(all.keySet());
            for (Iterator<String> i = keys.iterator();i.hasNext();) {
                String key = i.next();
                GainAnalysis value = all.get(key);
                for (int j=0;j<key.length();j++) {
                    char c = key.charAt(j);
                    if (c=='&') {
                        key = key.substring(0, j) + "&amp;" + key.substring(j+1);
                        j += 4;
                    } else if (c=='<') {
                        key = key.substring(0, j) + "&lt;" + key.substring(j+1);
                        j += 3;
                    } else if (c=='>') {
                        key = key.substring(0, j) + "&gt;" + key.substring(j+1);
                        j += 3;
                    }
                }
                writer.write("<track>\n <file>");
                writer.write(key);
                writer.write("</file>\n <analysis>");
                writer.write(value.toString());
                writer.write("</analysis>\n</track>\n");
            }
            writer.write("</normalization>\n");
            writer.close();

            log("Saved "+all.size()+" tracks in "+(System.currentTimeMillis() - lastsave)+"ms", null);
            if (!tempfile.renameTo(dbfile)) {
                throw new IOException("Cannot rename "+tempfile+" to "+dbfile);
            }
        }
    }

    /**
     * Set how many seconds apart the database should be automatically saved
     * @param seconds the number of seconds between saves, or 0 to never save.
     */
    public void setAutoSave(int seconds) {
        this.autosave = seconds <= 0 ? Long.MAX_VALUE : seconds * 1000;
    }

    /**
     * Set the Database file. The file is assumed to be absolute
     */
    public void setDatabase(File file) {
        if (all != null) {
            throw new IllegalStateException("Already initialized");
        }
        this.dbfile = file;
    }

    /**
     * Remove the previously calculated analysus for file from the database
     */
    public void remove(File file) throws IOException {
        load();
        String key = getKey(file);
        if (all.remove(key) != null) {
            if ((lastmodified=System.currentTimeMillis()) - lastsave > autosave) {
                try {
                    saveDatabase();
                } catch (IOException e) {
                    log("Can't remove \""+file+"\"", e);
                }
            }
        }
    }

    /**
     * Start an analysis on the specified file, which will add the
     * {@link GainAnalysis} to the database when it completes. If the
     * file already exists in the database it's entry will be overwritten.
     * If the file is a directory it's contents will be enqueued instead.
     */
    public void enqueue(File file) throws IOException {
        load();
        final File absfile = FileUtils.getAbsoluteFile(file);
        if (absfile.isDirectory()) {
            File[] list = absfile.listFiles();
            Arrays.sort(list);
            for (int i=0;i<list.length;i++) {
                enqueue(list[i]);
            }
        } else {
            while (executor.getQueue().size() > executor.getPoolSize() * 2) {
                try {
                    Thread.sleep(100);
                } catch(InterruptedException e) { }
            }
            executor.submit(new Runnable() {
                public void run() {
                    process(absfile);
                }
            });
        }
    }

    /**
     * Get the GainAnalysis for the specified file. If it doesn't
     * yet exists, it will be run immediately (no queueing) and returned.
     * If it can't be determined due to failure, return null.
     */
    public GainAnalysis getGain(File file) throws IOException {
        GainAnalysis analysis = null;
        file = FileUtils.getAbsoluteFile(file);
        String key = getKey(file);
        while (pending.containsKey(key) && !Thread.interrupted()) {
            synchronized(this) {
                try {
                    wait();
                } catch (InterruptedException e) {}
            }
        }
        analysis = all.get(key);
        if (analysis == null) {
            process(file);
            analysis = all.get(key);
        }
        return analysis.equals(GainAnalysis.FAILED) ? null : analysis;
    }

    private synchronized void load() {
        if (all == null) {
            lastsave = lastmodified = System.currentTimeMillis();
            all = new ConcurrentHashMap<String,GainAnalysis>();
            try {
                if (dbfile != null && dbfile.canRead()) {
                    InputStream in = new BufferedInputStream(new FileInputStream(dbfile));
                    XMLStreamReader r = XMLInputFactory.newFactory().createXMLStreamReader(in, "UTF-8");
                    int count = 0, state = 0;
                    String file = null;
                    for (int event = r.next();  event != XMLStreamConstants.END_DOCUMENT; event = r.next()) {
                        switch (event) {
                        case XMLStreamConstants.START_ELEMENT:
                            String tag = r.getLocalName();
                            if (tag.equals("normalization") && state == 0) {
                                state = 1;
                            } else if (tag.equals("track") && state == 1) {
                                state = 2;
                            } else if (tag.equals("file") && state == 2) {
                                file = java.text.Normalizer.normalize(r.getElementText(), java.text.Normalizer.Form.NFD);
                                state = 3;
                            } else if (tag.equals("analysis") && state == 3) {
                                GainAnalysis gain = GainAnalysis.fromString(r.getElementText());
                                all.put(file, gain);
                                count++;
                                state = 4;
                            } else {
                                throw new XMLStreamException("Invalid tag \""+tag+"\"");
                            }
                            break;
                        case XMLStreamConstants.END_ELEMENT:
                            if (r.getLocalName().equals("track")) {
                                state = 1;
                            }
                            break;
                        }
                    }
                    r.close();
                    in.close();
                    log("Loaded "+count+" tracks", null);
                }
            } catch (Exception e) {
                log("Error loading \""+dbfile+"\"", e);
            }
        }
    }

    private void process(File file) {
        if (file.canRead() && file.getName().toLowerCase().endsWith(".mp3")) {
            final String key = getKey(file);
            GainAnalysis analysis = all.get(key);
            if (analysis == null) {
                AudioInputStream in = null;
                try {
                    pending.put(key, "");
                    in = TrackReader.getTrackReader(file).getAudioInputStream(file);
                    log("Normalizing \""+file+"\"", null);
                    analysis = analyzer.get().analyze(in);
                    if (!Thread.interrupted()) {
                        all.put(key, analysis);
                        if ((lastmodified=System.currentTimeMillis()) - lastsave > autosave) {
                            saveDatabase();
                        }
                    }
                    synchronized(this) {
                        notifyAll();
                    }
                } catch (Exception  e) {
                    all.put(key, GainAnalysis.FAILED);
                    log("Can't normalize \""+file+"\"", e);
                } finally {
                    pending.remove(key);
                    if (in != null) {
                        try { in.close(); } catch (IOException e) { }
                    }
                }
            } else {
                log("Skipping \""+file+"\"", null);
            }
        }
    }

    private void log(String msg, Throwable e) {
        if (e != null) {
            out.println("WARNING: "+msg);
            e.printStackTrace(out);
        } else {
            if (debug) out.println("# "+msg);
        }
    }

    private String getKey(File file) {
        try {
            file = FileUtils.getRelativeFile(file);
            String s = java.text.Normalizer.normalize(file.getPath(), java.text.Normalizer.Form.NFD);
            for (int i=0;i<s.length();i++) {
                char c = s.charAt(i);
                if (c < 32) {
                    s = s.substring(0, i) + s.substring(i+1);
                    i--;
                }
            }
            return s;
        } catch (IOException e) {
            throw new IllegalStateException(e);
        }
    }

    /**
     * Wait untli all pending tasks are completed.
     * @return true if they're all done, false if we were interrupted or more were queued
     */
    public boolean waitForCompletion() {
        try {
            // Wait until no more files are pending...
            Thread.sleep(1);
            while (executor.getActiveCount() > 0) {
                Thread.sleep(1);
            }
        } catch (InterruptedException e) { }
        return executor.getActiveCount()==0;
    }

    /**
     * Shutdown the Normalizer. This must be called before the System will exit cleanly.
     */
    public void shutdown() throws InterruptedException {
        executor.shutdown();
        executor.awaitTermination(Long.MAX_VALUE, TimeUnit.SECONDS);
    }

}
