package org.qtunes.telnetd;

import org.qtunes.core.*;
import org.qtunes.auth.*;
import org.qtunes.auth.control.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.net.*;

/**
 * A Simple controller that uses the System.in and System.out to read
 * commands. It accepts "help" to list available commands, and the
 * commands are formatted as you would call a method in JavaScript,
 * eg <code>qtunes.quit()</code>, <code>AppleServer.addRemote("*", "8192")</code>
 * or even <code>db.getFields([1,2,3,4],["Artist","Title","Album"])</code>
 */
class TelnetConnection {

    private static final String NL = "\r\n";
    private InputStream in;
    private OutputStream out;
    private InetSocketAddress address;
    private Session session;
    private ControlMethodInvoker invoker;
    private TelnetController controller;
    private TelnetParser parser;

    TelnetConnection(InputStream in, OutputStream out, InetSocketAddress address, TelnetController controller) {
        this.in = in;
        this.out = out;
        this.address = address;
        this.invoker = new ControlMethodInvoker(controller.getContext());
        this.parser = new TelnetParser();
        this.controller = controller;
    }

    private void write(String s) throws IOException {
        byte[] b = s.getBytes("ISO-8859-1");
        out.write(b, 0, b.length);
        out.flush();
    }

    private void setEcho(boolean echo) throws IOException {
        // WILL=fb=251, WONT=fc=253, DO=fd=253, DONT=fe=254,
        if (echo) {
//            write("\u00ff\u00fb\u0001");
        } else {
//            write("\u00ff\u00fc\u0001");
        }
    }

    private String read() throws IOException {
        StringBuffer sb = new StringBuffer();
        do {
            int z = in.read();
            if (z < 0) return null;
            if ((z > 127 || z < 32) && z!=13 && z!=10) {
//                System.out.println("Read "+z);
            } else {
                sb.append((char)z);
            }
        } while (!sb.toString().endsWith(NL));
        return sb.toString().substring(0, sb.length()-2);
    }

    boolean initialize() throws IOException {
        int count = 0;
        do {
            write(controller.getContext().getSoftwareName()+NL);
            write("Login: ");
            String name = read();
            if (name==null) return false;
            write("Password: ");
            setEcho(false);
            String password = read();
            if (password==null) return false;
            setEcho(true);
            write(NL);
            session = controller.getSession(name, password, address);
        } while (session==Session.NONE && count++ < 3);
        Session.setCurrentSession(session);
        return count < 3;
    }

    boolean parseLine() throws IOException {
        write("qtunes> ");
        String s = read();
        if (s==null) {
            return false;
        } else if (s.equals("quit") || s.equals("exit")) {
            return false;
        } else if (s.equals("help")) {
           write(help(null)+NL);
        } else if (s.startsWith("help ")) {
           write(help(s.substring(5))+NL);
        } else if (s.trim().length() > 0) {
            try {
                parser.parse(s);
                String objectname = parser.getObjectName();
                List<Call> calls = parser.getMethodCalls();
                String response = parser.toString(invoker.invoke(controller.getContext(), objectname, calls));
                if (response!=null) {
                    write(response+NL);
                }
            } catch (Throwable e) {
                write(parser.toString(e)+NL);
            }
        }
        return true;
    }

    /**
     * Return a String of help messages for the specified Object name
     * @param objectname the Object name, or null for general help
     */
    private String help(String objectname) {
        Object object = null;
        if (objectname!=null) {
            try {
                parser.parse(objectname);
                object = invoker.invoke(controller.getContext(), parser.getObjectName(), parser.getMethodCalls());
            } catch (Throwable e) {
                e.printStackTrace();
            }
        }
        if (object==null) {
            StringBuffer sb = new StringBuffer();
            sb.append("Type help \"name\", where 'name' is one of: qtunes");
            String[] servicenames = controller.getContext().getServiceNames();
            for (int i=0;i<servicenames.length;i++) {
                Service service = controller.getContext().getService(Service.class, "name=\""+servicenames[i]+"\"");
                if (!invoker.getControlMethods(service).isEmpty()) {
                    sb.append(", "+servicenames[i]);
                }
            }
            return sb.toString();
        } else {
            List<Method> methods = invoker.getControlMethods(object);
            StringBuffer sb = new StringBuffer();
            for (int i=0;i<methods.size();i++) {
                ControlMethod annot = (ControlMethod)methods.get(i).getAnnotation(ControlMethod.class);
                Class[] paramtypes = methods.get(i).getParameterTypes();
                String[] params = annot.params();
                sb.append(objectname+"."+methods.get(i).getName()+"(");
                for (int j=0;j<params.length;j++) {
                    if (j > 0) {
                        sb.append(", ");
                    }
                    String classname = paramtypes[j].getName();
                    classname = classname.substring(classname.lastIndexOf(".")+1);
                    sb.append(classname+" "+params[j]);
                }
                sb.append(")\n    ");
                sb.append(annot.help()+"\n");
            }
            return sb.toString();
        }
    }

    void destroy() throws IOException {
        Session.getCurrentSession().expire();
    }

}
