package org.qtunes.web.spi.simple;

import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*;
import org.qtunes.web.*;
import org.qtunes.core.util.HttpParameters;

/** 
 * A very simple implementation of WebConnection that reads and writes
 * from an InputStream/OutputStream combo. This class is essentially
 * the protocol layer of a webservice. It's used by the SimpleServer.
 */
class SimpleConnection implements WebConnection {

    private final PushbackInputStream in;
    private final OutputStream out;
    private final InetSocketAddress remoteaddress;
    private final WebService webservice;

    private URI uri;
    private String method, protocolversion;
    private HttpParameters parameters;
    private Map<CIString,List<String>> requestheaders = new HashMap<CIString,List<String>>();
    private Map<CIString,List<String>> responseheaders = new HashMap<CIString,List<String>>();
    private int responsecode = -1;
    private boolean closeconnection;

    /**
     * Create a new SimpleConnection
     * @param in the Socket InputStream
     * @param out the Socket OutputStream
     * @param remote the remote IP address
     */
    SimpleConnection(InputStream in, OutputStream out, InetSocketAddress remote, WebService webservice) {
        this.in = new PushbackInputStream(in, 1);
        this.out = out;
        this.remoteaddress = remote;
        this.webservice = webservice;
    }

    private static int readeof(InputStream in) throws IOException {
        try {
            return in.read();
        } catch (SocketException e) {
            if (e.getMessage().equals("Socket closed")) {
                return -1;
            } else {
                throw e;
            }
        }
    }

    /**
     * Read a line from the InputStream as ISO-8859-1
     * @param firstline if false, headers can be continued onto the next line (RFC2616 2.2)
     * @return a String, not including the trailing LF or CRLF, or null for EOF
     */
    private String readLine(boolean firstline) throws IOException {
        StringBuffer sb = new StringBuffer();
        int c;
        while ((c=readeof(in))>=0) {
            if (c=='\r') {
                c = readeof(in);
                if (c!='\n') {
                    sb.append('\r');
                    if (c >= 0) {
                        sb.append((char)c);
                    }
                }
            }
            if (c=='\n') {
                c = firstline || sb.length()==0 ? -1 : readeof(in);
                if (!firstline && (c==' ' || c=='\t')) {
                    in.unread(c);
                } else {
                    if (c >= 0) in.unread(c);
                    return sb.toString();
                }
            } else {
                sb.append((char)c);
                if (c==' ' || c=='\t') {
                    while ((c=readeof(in))==' ' || c=='\t');
                    if (c >= 0) in.unread(c);
                }
            }
        }
        return sb.length()==0 ? null : sb.toString();
    }

    /**
     * Parse an HTTP request from the InputStream
     * @return true if the request was parsed, false on EOF
     * @throws IOException if the request was invalid
     */
    boolean parse() throws IOException {
        responsecode = -1;
        String line = readLine(true);
        if (line!=null) {
            try {
                String[] firstline = line.split(" ");
                if (firstline.length == 3 && (firstline[0].equals("GET") || firstline[0].equals("POST")) && firstline[2].equals("HTTP/0.9") || firstline[2].equals("HTTP/1.0") || firstline[2].equals("HTTP/1.1")) {
                    method = firstline[0];
                    protocolversion = firstline[2];
                    uri = new URI(firstline[1]);
                    while ((line=readLine(false))!=null && line.length() > 0) {
                        int ix = line.indexOf(':');
                        if (ix > 0) {
                            CIString key = new CIString(line.substring(0, ix));
                            String value = line.substring(ix+1).trim();
                            List<String> list = requestheaders.get(key);
                            if (list==null) {
                                requestheaders.put(key, list = new ArrayList<String>(1));
                            }
                            list.add(value);
                        } else {
                            throw new IllegalArgumentException("Unknown header line "+line);
                        }
                    }
                    parameters = new HttpParameters(uri, "ISO-8859-1");
                    if (!uri.isAbsolute() && requestheaders.containsKey(new CIString("Host"))) {
                        uri = new URI("http://"+getRequestHeader("Host")).resolve(uri);
                    }
                    closeconnection = "close".equals(getRequestHeader("Connection")) || !protocolversion.equals("HTTP/1.1");
                    return true;
                } else {
                   throw new IOException("Invalid protocol \""+line+"\"");
                }
            } catch (IOException e) {
                throw e;
            } catch (URISyntaxException e) {
                throw new IOException("Invalid URI", e);
            } catch (RuntimeException e) {
                throw new IOException("Unable to parse \""+line+"\"", e);
            }
        } else {
            return false;
        }
    }

    /**
     * Destroy the connection by closing the socket
     */
    public void destroy() throws IOException {
        in.close();
        out.close();
    }

    //-------------------------------------------------------------

    public InetSocketAddress getRemoteAddress() {
        return remoteaddress;
    }

    public URI getRequestURI() {
        return uri;
    }

    public int getResponseCode() {
        return responsecode;
    }

    public void close() throws IOException {
        out.flush();
    }

    public void setResponseHeader(String key, Date value) {
        setResponseHeader(key, new SimpleDateFormat(WebHandler.DATEFORMAT).format(value));
    }

    public void addResponseHeader(String key, String value) {
        if (value!=null) {
            CIString keystring = new CIString(key);
            List<String> list = responseheaders.get(keystring);
            if (list==null) {
                responseheaders.put(keystring, list = new ArrayList<String>(1));
            }
            list.add(value);
        }
    }

    public void setResponseHeader(String key, String value) {
        CIString keystring = new CIString(key);
        if (value==null) {
            responseheaders.remove(keystring);
        } else {
            List<String> list = responseheaders.get(keystring);
            if (list==null) {
                responseheaders.put(keystring, list = new ArrayList<String>(1));
            }
            list.clear();
            list.add(value);
        }
    }

    public Collection<String> getRequestHeaderNames() {
        List<String> list = new ArrayList<String>(requestheaders.size());
        for (Iterator<CIString> i = requestheaders.keySet().iterator();i.hasNext();) {
            list.add(i.next().toString());
        }
        return list;
    }

    private String getMessage(int status) {
        switch (status) {
          case 200:
            return "OK";
          case 204:
            return "No Content";
          case 206:
            return "Partial Content";
          case 404:
            return "Not Found";
          case 500:
            return "Internal Error";
          case 401:
            return "Unauthorized";
          case 301:
            return "Moved permanently";
          case 302:
            return "Found";
          default:
            return "Message "+status;
        }
    }

    public void sendResponseHeaders(int status, int length) throws IOException {
        if (responsecode > 0) {
            throw new IllegalStateException("Headers already sent");
        }
        if (length < 0) {
            setResponseHeader("Content-Length", "0");
        } else if (length==0) {
            setResponseHeader("Content-Length", (String)null);
        } else {
            setResponseHeader("Content-Length", Integer.toString(length));
        }
        OutputStreamWriter writer = new OutputStreamWriter(out, "ISO-8859-1");
        writer.write(protocolversion+" "+Integer.toString(status)+" "+getMessage(status)+"\r\n");
        for (Iterator<Map.Entry<CIString,List<String>>> i = responseheaders.entrySet().iterator();i.hasNext();) {
            Map.Entry<CIString,List<String>> e = i.next();
            List<String> values = e.getValue();
            for (int j=0;j<values.size();j++) {
                writer.write(e.getKey()+": "+values.get(j)+"\r\n");
            }
        }
        writer.write("\r\n");
        writer.flush();
        responsecode = status;
    }

    public String getRequestHeader(String key) {
        List<String> list = requestheaders.get(new CIString(key));
        return list==null ? null : list.get(0);
    }

    public List<String> getRequestHeaders(String key) {
        return Collections.unmodifiableList(requestheaders.get(new CIString(key)));
    }

    public String getMethod() {
        return method;
    }

    public InputStream getInputStream() {
        if (responsecode < 0) {
            return in;
        } else {
            throw new IllegalStateException("Headers have been written");
        }
    }

    public OutputStream getOutputStream() {
        if (responsecode < 0) {
            throw new IllegalStateException("Headers have not been written");
        }
        if (out==null) {
            throw new IllegalStateException("OutputStream has been closed");
        }
        List<String> length = responseheaders.get(new CIString("Content-Length"));
        if (length!=null) {
            final int len = Integer.parseInt(length.get(0));
            return new FilterOutputStream(out) {
                int count;
                public void write(int b) throws IOException {
                    if (++count > len) throw new IOException("Output > "+len);
                    out.write(b);
                }
                public void write(byte[] b) throws IOException {
                   write(b, 0, b.length);
                }
                public void write(byte[] b, int off, int buflen) throws IOException {
                    if (count+buflen > len) throw new IOException("Output > "+len);
                    out.write(b, off, buflen);
                    count += buflen;
                }
                public void close() throws IOException {
                    if (count!=len) throw new IOException("Output len "+count+" < "+len);
                    if (closeconnection) {
                        super.close();
                    }
                }
            };
        } else {
            return out;
        }
    }

    public String getParameter(String key) {
        return parameters.getValue(key);
    }

    public List<String> getParameterValues(String key) {
        return parameters.getValues(key);
    }

    public Map<String,Object> getParameters() {
        return parameters.getAllValues();
    }

    public WebService getWebService() {
        return webservice;
    }

    //-------------------------------------------------------------------------

    private static class CIString implements Comparable {
        private final String string;
        CIString(String string) {
            this.string = string;
        }
        public String toString() {
            return string;
        }
        public int hashCode() {
            return string.toLowerCase().hashCode();
        }
        public boolean equals(Object o) {
            return string.equalsIgnoreCase(((CIString)o).string);
        }
        public int compareTo(Object o) {
            return string.toLowerCase().compareTo(((CIString)o).string.toLowerCase());
        }
    }

}
