package org.qtunes.webplayer;

import org.qtunes.core.*;
import org.qtunes.auth.*;
import org.qtunes.player.*;
import org.qtunes.db.*;
import org.qtunes.auth.control.*;
import org.qtunes.web.*;
import org.qtunes.core.util.Base64;
import org.qtunes.core.util.JSONSerializer;
import org.qtunes.core.util.HttpParameters;
import java.net.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

/**
 * Returns the table data for the main track listing
 */
class GridHandler implements WebHandler {
   
    private final ServiceContext context;
    private final Database db;
    private JSONSerializer json;
    private String dateformat;
    private List<AbstractField> fields;
    private Map<AbstractField,String> fieldformats;

    GridHandler(WebPlayer player) {
        this.context = player.getContext();
        this.db = context.getService(Database.class);
        this.json = new JSONSerializer();
        dateformat = context.getProperty("dateformat");
        if (dateformat == null) {
            dateformat = "dd MMM yyyy HH:mm:ss";
        }
        String fieldnames = context.getProperty("fields");
        if (fieldnames == null) {
            fieldnames = "Title, Artist, AlbumArtist, Album, Track, Disc, Year, Group, LastModified";
        }
        String[] z = fieldnames.split(",");
        fields = new ArrayList<AbstractField>();
        fieldformats = new HashMap<AbstractField,String>();
        for (int i=0;i<z.length;i++) {
            AbstractField field = Field.getByName(z[i].trim());
            if (field != null && !fields.contains(field)) {
                fields.add(field);
                String format = context.getProperty("format."+field.getName());
                if (format != null) {
                    fieldformats.put(field, format);
                }
            }
        }
    }

    private static String getSortKey(String sort, String dir) {
        dir = "DESC".equals(dir) ? "-" : "+";
        if (sort == null || sort.equalsIgnoreCase("n")) {
            return null;
        } else if (sort.equalsIgnoreCase("Title")) {
            return dir+"title";
        } else if (sort.equalsIgnoreCase("Album")) {
            return dir+"album";
        } else if (sort.equalsIgnoreCase("Artist")) {
            return dir+"artist";
        } else if (sort.equalsIgnoreCase("AlbumArtist")) {
            return dir+"albumartist";
        } else if (sort.equalsIgnoreCase("Group")) {
            return dir+"group";
        } else {
            throw new IllegalStateException("Unknown sort field \""+sort+"\"");
        }
    }

    private Playlist getPlaylist(String name) {
        if (name.charAt(0)=='#') {
            return context.getService(Player.class, "name=\""+name.substring(1)+"\"").getPlaylist();
        } else {
            return db.getPlaylist(name);
        }
    }

    private HttpParameters parsePostParameters(WebConnection con) throws IOException {
        BufferedReader reader = new BufferedReader(new InputStreamReader(con.getInputStream(), "UTF-8"));
        Map<String,Object> params = new HashMap<String,Object>();
        String line;
        StringBuilder sb = new StringBuilder();
        while ((line=reader.readLine())!=null && line.length() > 0) {
            sb.append(line);
        }
        reader.close();
        return new HttpParameters(sb.toString(), "UTF-8");
    }

    @SuppressWarnings("unchecked")
    public void handle(WebConnection con) throws IOException {
        OutputStream out = null;
        String playlistname = null, sort = null, dir = null, filter = null;
        int start = 0, limit = 0;
        DateFormat df = new SimpleDateFormat(dateformat);
        HttpParameters params = con.getMethod().equals("POST") ? parsePostParameters(con) : new HttpParameters(con.getRequestURI().getRawQuery(), "ISO-8859-1");

        playlistname = params.getValue("playlist");
        if (playlistname == null) {
            throw new IOException("No playlist "+con.getRequestURI());
        }
        sort = params.getValue("sort");
        filter = params.getValue("filter");
        dir = params.getValue("dir");
        try {
            start = Integer.parseInt(params.getValue("start"));
        } catch (Exception e) { }
        try {
            limit = Integer.parseInt(params.getValue("limit"));
        } catch (Exception e) { }
        con.setResponseHeader("Content-Type", "application/json;charset=UTF-8");


        List<Track> sortedtracks, unsortedtracks;
        Playlist unsortedplaylist = getPlaylist(playlistname);
        Playlist sortedplaylist;
        if (unsortedplaylist == null) {
            sortedplaylist = null;
            sortedtracks = unsortedtracks = new ArrayList<Track>();
        } else {
            sortedplaylist = unsortedplaylist.sort(getSortKey(sort, dir), filter);
            try {
                db.lock(true);
                unsortedtracks = unsortedplaylist.getTracks();
                sortedtracks = sortedplaylist.getTracks();
            } finally {
                db.lock(false);
            }
        }
        unsortedtracks = new ArrayList<Track>(unsortedtracks);  // We'll be editing

        List<Map<String,Object>> list = new ArrayList<Map<String,Object>>(limit);
        if (start + limit > sortedtracks.size()) {
            limit = sortedtracks.size() - start;
        }
        for (int i=start;i<start+limit;i++) {
            final Track track = sortedtracks.get(i);
            Map<String,Object> vals = new LinkedHashMap<String,Object>();
            vals.put("id", Integer.valueOf(track.getIndex()));
            int ix = unsortedtracks.indexOf(track);
            vals.put("n", new Integer(ix+1));
            if (ix >= 0) {      // Shouldn't happen, but
                unsortedtracks.set(ix, null);
            } else {
                context.debug("Unsorted != Sorted: sorted="+sortedplaylist+" unsorted="+unsortedplaylist+" i="+i+" track="+track);
            }
            for (int j=0;j<fields.size();j++) {
                AbstractField field = fields.get(j);
                Object val = track.get(field);
                if (val != null) {
                    val = formatField(field, val, fieldformats.get(field), track, df);
                    vals.put(field.getName(), val);
                }
            }
            list.add(vals);
        }

        Map<String,Object> map = new LinkedHashMap<String,Object>();
        map.put("data", list);
        map.put("version", 1);
        map.put("count", sortedtracks.size());
        String response = json.write(map);
        byte[] data = response.getBytes("UTF-8");
        con.sendResponseHeaders(200, data.length);
        out = con.getOutputStream();
        out.write(data);
        out.close();
    }

    @SuppressWarnings("unchecked")
    private static Object formatField(AbstractField field, Object value, String format, Track track, DateFormat df) {
        if (format == null) {
            if (value == Integer.valueOf(0)) {
                value = null;
            } else if (field == Field.LastModified) {
                value = df.format(new Date(((Integer)value).intValue() * 1000l));
            }
            return value;
        }
        StringBuilder sb = new StringBuilder();
        for (int i=0;i<format.length();i++) {
            char c = format.charAt(i);
            if (c == '$' && i < format.length() - 1 && format.charAt(i+1) == '{') {
                int j;
                for (j=i+2;j<format.length() && format.charAt(j)!='}';j++);
                if (j != format.length()) {
                    String s = format.substring(i+2, j), prefix = "";
                    i = j;
                    j = s.indexOf(':');
                    if (j > 0) {
                        prefix = s.substring(j+1);
                        s = s.substring(0, j);
                    }
                    field = Field.getByName(s);
                    if (field != null) {
                        value = track.get(field);
                        value = formatField(field, value, null, track, df);
                        if (value != null) {
                            sb.append(prefix);
                            sb.append(value);
                        }
                    }
                    c = 0;
                }
            }
            if (c != 0) {
                sb.append(c);
            }
        }
        return sb.toString();
    }
}
