package org.qtunes.webplayer;

import org.qtunes.core.*;
import org.qtunes.auth.*;
import org.qtunes.daap.DaapServer;
import org.qtunes.db.*;
import org.qtunes.player.*;
import org.qtunes.speaker.*;
import org.qtunes.auth.control.*;
import org.qtunes.core.util.HttpParameters;
import org.qtunes.web.*;
import java.net.*;
import java.io.*;
import java.util.*;

/**
 * A controller which runs the actions specified by the web interface
 */
class WebController implements WebHandler {
   
    private ServiceContext context;
    private String prefix;
    private final Auth auth;
    private final JSONParser jsonparser;

    public WebController(WebPlayer player, String prefix, WebAuthorizer authorizer) {
        this.context = player.getContext();
        if (!prefix.endsWith("/")) {
            prefix += "/";
        }
        this.prefix = prefix;
        auth = context.getService(Auth.class);
        jsonparser = new JSONParser(context);   // For writing only
    }

    private Player getPlayer(HttpParameters params) {
        return params.getValue("player")==null ? null : context.getService(Player.class, "name=\""+params.getValue("player")+"\"");
    }

    private Speaker getSpeaker(HttpParameters params) {
        return params.getValue("speaker")==null ? null : context.getService(Speaker.class, "name=\""+params.getValue("speaker")+"\"");
    }

    private Playlist getPlaylist(HttpParameters params, String key) {
        if (key == null) {
            key = "playlist";
        }
        Playlist playlist = context.getService(Database.class).getPlaylist(params.getValue(key));
        String sortfield = params.getValue("sortfield");
        String sortdir = params.getValue("sortdir");
        String filter = params.getValue("filter");
        if ("".equals(filter)) {
            filter = null;
        }
        if ("".equals(sortfield) || "n".equals(sortfield)) {
            sortfield = null;
        }
        if (sortfield != null || filter != null) {
            sortfield = ("DESC".equals(sortdir) ? "-" : "+") + sortfield;
            playlist = playlist.sort(sortfield, filter);
        }
        return playlist;
    }

    private void authorize(String action) {
        Session session = Session.getCurrentSession();
        if (session.isAllowed(action)) {
            return;
        }
        throw new IllegalStateException("Not authorized for \""+action+"\"");
    }

    private ManualPlaylist getManualPlaylist(HttpParameters params) {
        Database db = context.getService(Database.class);
        ManualPlaylist pl = db.createManualPlaylist();
        String[] s = params.getValue("tracks").split(",");
        for (int i=0;i<s.length;i++) {
            Track track = db.getTrack(Integer.parseInt(s[i]));
            if (track != null) {
                pl.add(track);
            }
        }
        return pl;
    }

    private BitSet getIndices(HttpParameters params) {
        String[] s = params.getValue("indices").split(",");
        BitSet b = new BitSet();
        for (int i=0;i<s.length;i++) {
            b.set(Integer.parseInt(s[i]));
        }
        return b;
    }

    public void handle(WebConnection con) throws IOException {
        con.setResponseHeader("Server", context.getSoftwareName()+"/"+con.getWebService().getContext().getServiceName());
        con.setResponseHeader("Date", new Date());
        URI uri = con.getRequestURI();
        String response = "{ok:true}";
        try {
            String in = uri.getRawPath();
            if (in.startsWith(prefix)) {
                in = in.substring(prefix.length());
            }
            context.debug("IN="+in+" Q="+uri.getRawQuery());
            try {
                HttpParameters params;
                if (uri.getRawQuery()!=null) {
                    params = new HttpParameters(uri.getRawQuery(), "ISO-8859-1");
                } else {
                    params = new HttpParameters("", "ISO-8859-1");
                }
                if (params.containsKey("session")) {
                    String val = params.getValue("session");
                    Session.setCurrentSession(auth.getSession(Integer.parseInt(val)));
                    params.remove("session");
                }

                Player player = getPlayer(params);
                if (in.equals("login")) {
                    Map<String,Object> args = params.getAllValues();
                    args.put("ip", con.getRemoteAddress());
                    String useragent = con.getRequestHeader("User-Agent");
                    if (useragent!=null) {
                        args.put("user-agent", useragent);
                    }
                    String referer = con.getRequestHeader("Referer");
                    if (referer!=null) {
                        args.put("referer", referer);
                    }
                    String host = con.getRequestHeader("Host");
                    if (host!=null) {
                        args.put("host", useragent);
                    }
                    args.put("service", context.getService());

                    response = jsonparser.toString(auth.login(args));
                } else if (in.equals("play") || in.equals("enqueue")) {
                    authorize("play");
                    Playlist playlist;
                    if (params.containsKey("playlist")) {
                        playlist = getPlaylist(params, null);
                    } else {
                        playlist = getManualPlaylist(params);
                    }
                    if (in.equals("play")) {
                        player.getPlaylist().clear();
                        player.setPaused(false);
                    }
                    player.getPlaylist().addAll(playlist);
                } else if (in.equals("dequeue")) {
                    authorize("play");
                    player.getPlaylist().removeAll(getIndices(params));
                } else if (in.equals("clear")) {
                    authorize("play");
                    player.setPaused(false);
                    player.getPlaylist().clear();
                } else if (in.equals("editplaylist")) {
                    authorize("play");
                    Database db = context.getService(Database.class);
                    String playlistname = params.getValue("playlist");
                    String query = params.getValue("query");
                    ((SmartPlaylist)db.getPlaylist(playlistname)).setQuery(query);
                } else if (in.equals("createplaylist")) {
                    authorize("play");
                    Database db = context.getService(Database.class);
                    String playlistname = params.getValue("playlist");
                    if (params.containsKey("tracks")) {
                        db.putPlaylist(playlistname, getManualPlaylist(params));
                    } else {
                        db.putPlaylist(playlistname, db.createSmartPlaylist(params.getValue("query")));
                    }
                } else if (in.equals("deleteplaylist")) {
                    authorize("play");
                    Database db = context.getService(Database.class);
                    db.putPlaylist(params.getValue("playlist"), null);
                } else if (in.equals("renameplaylist")) {
                    authorize("play");
                    Database db = context.getService(Database.class);
                    Playlist pl = getPlaylist(params, "from");
                    db.putPlaylist(params.getValue("to"), pl);
                } else if (in.equals("setvolume")) {
                    authorize("play");
                    if (player!=null) {
                        player.setVolume(Integer.parseInt(params.getValue("volume")));
                    } else {
                        Speaker speaker = getSpeaker(params);
                        if (speaker!=null) {
                            speaker.setVolumeAdjustment(Integer.parseInt(params.getValue("volume")));
                            if (speaker.getPlayer()!=null) {
                                speaker.getPlayer().setVolume(speaker.getPlayer().getVolume());
                            }
                        }
                    }
                } else if (in.equals("seek")) {
                    authorize("play");
                    player.seek(Integer.parseInt(params.getValue("ms")));
                } else if (in.equals("addtracks")) {
                    authorize("play");
                    ((ManualPlaylist)getPlaylist(params, null)).addAll(getManualPlaylist(params));
                } else if (in.equals("deletetracks")) {
                    authorize("play");
                    ((ManualPlaylist)getPlaylist(params, null)).removeAll(getIndices(params));
                } else if (in.equals("pause")) {
                    authorize("play");
                    if (player.getCurrentTrack() == null) {
                        player.setPaused(false);
                        player.setCurrentTrackIndex(0);
                    } else {
                        player.setPaused("1".equals(params.getValue("paused")));
                    }
                } else if (in.equals("playnext")) {
                    authorize("play");
                    player.setPaused(false);
                    player.playNext();
                } else if (in.equals("playprevious")) {
                    authorize("play");
                    player.setPaused(false);
                    player.playPrevious();
                } else if (in.equals("setshuffle")) {
                    authorize("play");
                    player.setShuffle("true".equals(params.getValue("shuffle")));
                } else if (in.equals("setrepeat")) {
                    authorize("play");
                    player.setRepeat(Integer.parseInt(params.getValue("repeat")));
                } else if (in.equals("setplayindex")) {
                    authorize("play");
                    player.setPaused(false);
                    player.setCurrentTrackIndex(Integer.parseInt(params.getValue("index")));
                } else if (in.equals("authorizeremote")) {
                    authorize("play");
                    DaapServer server = context.getService(DaapServer.class);
                    server.addRemote(params.getValue("remote"), player, params.getValue("pin"));
                } else if (in.equals("setspeakers")) {
                    authorize("play");
                    for (Iterator<String> i = params.getAllValues().keySet().iterator();i.hasNext();) {
                        String speakername = i.next();
                        String playername = params.getValue(speakername);
                        Speaker speaker = context.getService(Speaker.class, "name=\""+speakername+"\"");
                        if (speaker != null) {
                            player = playername.length()==0 ? null : context.getService(Player.class, "name=\""+playername+"\"");
                            speaker.setPlayer(player);
                        }
                    }
                } else {
                    throw new IOException("Unknown action \""+in+"\"");
                }

            } catch (Throwable e) {
                context.debug(in, e);
                response = jsonparser.toString(e);
            } finally {
                Session.setCurrentSession(null);
            }

            con.addResponseHeader("Content-Type", "application/json;encoding=UTF-8");
            byte[] data = response.getBytes("UTF-8");
            con.sendResponseHeaders(200, data.length);
            OutputStream out = con.getOutputStream();
            out.write(data);
            out.close();
        } catch (Throwable e) {
            context.debug(null, e);
            if (!(e instanceof IOException && "Broken pipe".equals(e.getMessage()))) {
                context.error(uri.toString(), e);
            }
            if (e instanceof IOException) {
                throw (IOException)e;
            } else {
                throw new IOException(e);
            }
        }
    }

}
