package org.qtunes.zeroconf.spi.apple;

import org.qtunes.core.*;
import org.qtunes.zeroconf.*;
import com.apple.dnssd.*;
import java.util.*;

public class AppleZeroConf implements ZeroConf, RegisterListener {

    private ServiceContext context;

    public ZCService register(String type, String name, int port, Map<String,String> properties) {
        try {
            TXTRecord txt = null;
            if (properties!=null) {
                txt = new TXTRecord();
                for (Iterator<Map.Entry<String,String>> i = properties.entrySet().iterator();i.hasNext();) {
                    Map.Entry<String,String> e = i.next();
                    txt.set(e.getKey(), e.getValue());
                }
            }
            DNSSDRegistration registration = DNSSD.register(0, 0, name, type, null, null, port, txt, this);
            return new AppleService(registration);
        } catch (DNSSDException e) {
            throw new RuntimeException(e);
        }
    }

    public void unregister(ZCService service) {
        ((AppleService)service).wrapped.stop();
    }

    public ZCServiceInfo[] list(String type, int ms) {
        BrowseManager manager = new BrowseManager(type);
        return manager.list(ms);
    }

    private static class AppleServiceInfo implements ZCServiceInfo {
        final String name, type, host;
        final int port;
        final Map<String,String> properties;
        AppleServiceInfo(String type, String name, String host, int port, Map<String,String> properties) {
            this.type = type;
            this.name = name;
            this.host = host;
            this.port = port;
            this.properties = properties;
        }

        public int hashCode() {
            return type.hashCode() ^ name.hashCode() ^ port ^ (host==null ? 0 : host.hashCode());
        }

        public boolean equals(Object o) {
            if (o instanceof AppleServiceInfo) {
                AppleServiceInfo s = (AppleServiceInfo)o;
                return type.equals(s.type) && port==s.port && name.equals(s.name) && ((host==null && s.host==null) || (host!=null && host.equals(s.host)));
            } else {
                return false;
            }
        }

        public String getType() {
            return type;
        }

        public String getName() {
            return name;
        }

        public String getHost() {
            return host;
        }

        public int getPort() {
            return port;
        }

        public Map<String,String> getProperties() {
            return properties;
        }
    }

    private static class AppleService implements ZCService {
        private DNSSDRegistration wrapped;
        AppleService(DNSSDRegistration reg) {
            this.wrapped = reg;
        }
    }

    private static class BrowseManager implements BrowseListener, ResolveListener {
        final String type;
        DNSSDService browseservice;
        Set<ZCServiceInfo> results = new LinkedHashSet<ZCServiceInfo>();
        Set<DNSSDService> pending = Collections.synchronizedSet(new HashSet<DNSSDService>());
        boolean morecoming;

        BrowseManager(String type) {
            this.type = type;
        }

        ZCServiceInfo[] list(int ms) {
            try {
                long end = ms <= 0 ? Long.MAX_VALUE : System.currentTimeMillis() + ms;
                browseservice = DNSSD.browse(type, this);
                synchronized(this) {
                    do {
                        try {
                            wait(end - System.currentTimeMillis());
                        } catch (InterruptedException e) { }
                    } while ((!pending.isEmpty() || morecoming) && System.currentTimeMillis() < end);
                }
                browseservice.stop();
                return (ZCServiceInfo[])results.toArray(new ZCServiceInfo[0]);
            } catch (DNSSDException e) {
                throw new RuntimeException(e);
            }
        }

        public synchronized void serviceFound(DNSSDService browser, int flags, int ifIndex, String name, String type, String domain) {
            morecoming = (flags & DNSSD.MORE_COMING)!=0;
            try {
                pending.add(DNSSD.resolve(flags, ifIndex, name, type, domain, this));
            } catch (DNSSDException e) {
                throw new RuntimeException(e);
            }
        }

        public void serviceLost(DNSSDService browser, int flags, int ifIndex, String name, String type, String domain) {
            morecoming = (flags & DNSSD.MORE_COMING)!=0;
        }

        public synchronized void serviceResolved(DNSSDService resolver, int flags, int ifIndex, String name, String host, int port, TXTRecord txt) {
            resolver.stop();
            Map<String,String> properties = new LinkedHashMap<String,String>();
            for (int i=0;i<txt.size();i++) {
                if (txt.getKey(i).length() > 0) {
                    properties.put(txt.getKey(i), txt.getValueAsString(i));
                }
            }
            int ix = name.lastIndexOf("."+type);
            name = name.substring(0, ix);
            results.add(new AppleServiceInfo(type, name, host, port, properties));
            pending.remove(resolver);
            notifyAll();
        }

        public synchronized void operationFailed(DNSSDService service, int error) {
            pending.remove(service);
            notifyAll();
        }
    }

    public void serviceRegistered(DNSSDRegistration registration, int flags, String serviceName, String regType, String domain) {
    }

    public synchronized void operationFailed(DNSSDService service, int error) {
    }

    //---------------------------------------------------------------------------

    public ServiceContext getContext() {
        return context;
    }

    public void startService(ServiceContext context) {
        this.context = context;
    }

    public void stopService(ServiceContext context) {
    }

    public Map<String,Object> reportState() {
        return null;
    }

}
